//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains an example implementation of the SABR stochastic
// volatility model.
// See MANAGING SMILE RISK, 2002, Hagan P, Kumar D, Lesniekski A, Woodward D.
//============================================================================
//============================================================================
#include <windows.h>
#include <math.h>
#include "XllAddIn.h"
#include "sabr.h"

//============================================================
#define SABR_CALC_ALPHA_MAX_NR_ITERS	10

bool sabr_calc_alpha(double FwdRate, double AtmVol, double TexpYrs,
	double Beta, double VolVol, double Rho, double &Alpha,
	bool is_normal)
{
	double a0, a1, a2, a3, b1, b2, b3;
	double pow_f = pow(FwdRate, Beta - 1.0);
	double var = pow_f * pow_f * TexpYrs; // to simplify calculations

	a0 = -AtmVol;
	a1 = pow_f * (1.0 + TexpYrs * (2.0 - 3.0 * Rho * Rho) * VolVol * VolVol / 24.0);
	a2 = var * Rho * Beta * VolVol / 4.0;
	a3 = var * pow_f / 24.0; // not the final value

	if(is_normal)
	{
		a1 *= FwdRate;
		a2 *= FwdRate;
		a3 *= FwdRate * Beta * (Beta - 2.0);
		Alpha = 0.01; // Rough first guess = 1%
	}
	else
	{
		a3 *= (Beta - 1.0) * (Beta - 1.0);
		Alpha = AtmVol / a1; // Reasonable first guess
	}

// Calculate coefficients of 1st derivative polynomial
	b1 = a1;
	b2 = 2.0 * a2;
	b3 = 3.0 * a3;

// Run a Newton-Raphson search for the root, Alpha
	double correction;
	for(short i = SABR_CALC_ALPHA_MAX_NR_ITERS; i--;)
	{
		correction = (a0 + Alpha * (a1 + Alpha * (a2 + Alpha * a3)))
			/ (b1 + Alpha * (b2 + Alpha * b3));
		Alpha -= correction;
		if(fabs(correction) <= 1e-12)
			return true;
	}

// Should have converged by now but didn't, so should really
// implement a fall-back scheme here.  Instead, just fail.
	return false;
}
//============================================================
// XLL interface function for sabr_calc_alpha()
//============================================================
xloper * __stdcall SabrCalcAlpha(double Texp, double AtmVol,
		double Fwd, double Beta, double VolVol, double Rho,
		xloper *p_is_normal)
{
#pragma EXPORT
// Check inputs:
	if(Texp <= 0.0 || AtmVol <= 0.0 || Fwd <= 0.0
	|| Beta < 0.0 || Beta > 1.0 // Could relax upper limit on Beta
	|| VolVol < 0.0 // Allow zero: non-stochastic case
	|| Rho < -1.0 || Rho > 1.0)
		return p_xlErrValue;

	cpp_xloper IsNormal(p_is_normal);
	double Alpha;
	if(!sabr_calc_alpha(Fwd, AtmVol, Texp, Beta, VolVol,
		Rho, Alpha, IsNormal.IsTrue()))
		return p_xlErrNum;

	cpp_xloper RetVal(Alpha);
	return RetVal.ExtractXloper();
}
//============================================================
//============================================================
//============================================================
//============================================================
//============================================================
// Calc skewed BS vol from SABR params, expiry, fwd and strike
//============================================================
#define PHI_(y)		(1.0 + (y) / 24.0 * (1.0 + (y) / 80.0))

#define PHI(y)		(1.0 + (y) / 24.0 * (1.0 + (y) / 80.0 * (1.0 + (y) / 168.0)))

bool sabr_vol(double FwdRate, double Strike, double Texp,
			  double Alpha, double Beta, double VolVol,
			  double Rho, double &Vol, bool is_normal)
{
	double A = Alpha * pow(FwdRate * Strike, (Beta - 1.0) / 2.0); // > 0
	double h = log(FwdRate / Strike); // Strike always > 0
	double z = VolVol * h / A; // A always > 0
	double H = (Beta - 1.0) * (Beta - 1.0) * (h *= h); // h is squared
	double Phi_H = PHI(H);
	double P, lambda, z_by_x;

	if(is_normal)
	{
		P = FwdRate * Strike * PHI(h);
		lambda = Beta * (Beta - 2.0);
	}
	else // Lognormal (Black) vol
	{
		P = 1.0;
		lambda = (Beta - 1.0) * (Beta - 1.0);
	}

	if(fabs(z) < 1e-9)
	{
		z_by_x = sqrt(1.0 - Rho * z);
	}
	else
	{
		if(Rho == 1.0)
		{
			if(z >= 1.0)
				return false;

			z_by_x = z / -log(1.0 - z);
		}
		else
		{
			z_by_x = z /
				log((sqrt(1.0 + z * (z - 2.0 * Rho)) + z - Rho)
					/ (1.0 - Rho));
		}
	}
	Vol = A * P / Phi_H * z_by_x 
		* (1.0 + (A / 4.0 * (A * lambda / 6.0 + Rho * Beta * VolVol)
		+ (2.0 - 3.0 * Rho * Rho) * VolVol * VolVol / 24.0) * Texp);

	return true;
}
//============================================================
xloper * __stdcall SabrCalcVol(double Texp, double Fwd, double Strike,
	double Alpha, double Beta, double VolVol, double Rho, xloper *p_is_normal)
{
#pragma EXPORT

// Check inputs:
	if(Texp <= 0.0 || Fwd <= 0.0 || Strike <= 0.0
	|| Alpha <= 0.0
	|| Beta < 0.0 || Beta > 1.0 // Could relax upper limit on Beta
	|| VolVol < 0.0 // Allow zero: non-stochastic case
	|| Rho < -1.0 || Rho > 1.0)
		return p_xlErrValue;

	cpp_xloper IsNormal(p_is_normal);
	double Vol;
	if(!sabr_vol(Fwd, Strike, Texp, Alpha, Beta, VolVol, Rho,
		Vol, IsNormal.IsTrue()))
		return p_xlErrNum;

	cpp_xloper RetVal(Vol);
	return RetVal.ExtractXloper();
}
//============================================================
//============================================================
//============================================================
bool SabrWrapper::SetStrikeIndependents(double fwd_rate)
{
	m_FwdRate = fwd_rate;
	m_lambda_B = (m_Beta - 1.0) * (m_Beta - 1.0);
	m_lambda_N = m_lambda_B - 1.0;
	m_a_B = m_lambda_B / 24.0;
	m_a_N = m_lambda_N / 24.0;
	m_b = m_Rho * m_Beta * m_VolVol / 4.0;
	m_c = (2.0 - 3.0 * m_Rho * m_Rho) / 24.0 * m_VolVol * m_VolVol;
	m_d = (m_Beta - 1.0) / 2.0;
	return true;
}
//============================================================
bool SabrWrapper::GetVolOfStrike(double Strike, double &Vol, bool is_normal)
{
	double A = m_Alpha * pow(m_FwdRate * Strike, m_d); // > 0
	double h = log(m_FwdRate / Strike); // Strike always > 0
	double z = m_VolVol * h / A; // A always > 0
	double H = m_lambda_B * (h *= h); // h is squared
	double Phi_H = PHI(H);
	double P, lambda, z_by_x, a;

	if(is_normal)
	{
		P = m_FwdRate * Strike * PHI(h);
		lambda = m_lambda_N;
		a = m_a_N;
	}
	else // Lognormal (Black) vol
	{
		P = 1.0;
		lambda = m_lambda_N;
		a = m_a_B;
	}

	if(fabs(z) < 1e-9)
	{
		z_by_x = sqrt(1.0 - m_Rho * z);
	}
	else
	{
		if(m_Rho == 1.0)
		{
			if(z >= 1.0)	return false;
			z_by_x = z / -log(1.0 - z);
		}
		else
		{
			z_by_x = z / log((sqrt(1.0 + z * (z - 2.0 * m_Rho)) + z - m_Rho) / (1.0 - m_Rho));
		}
	}
	Vol = A * P / Phi_H * z_by_x * (1.0 + (A * (A * a + m_b) + m_c) * m_Texp);
	return true;
}
//============================================================


SabrWrapper SabrTest;


//============================================================
FnRegData RT_SabrTestSetStrikeInds( // Instance name is not important
	"SabrTestSetStrikeInds",	// Function name as in code (v11-)
	"BBBBBBB",		// Return type and argument types (v11-)
	"",				// Function name as in code (v12+)
	"",				// Return type and argument types (v12+)
	"SabrTestSetStrikeInds",	// Fn name as it appears on worksheet
	"Texp,Fwd,Alpha,Beta,VolVol,Rho",		// Argument names (Excel 11-: max 30)
	"",				// Argument names (Excel 12+: max 255)
	"Example",		// Function category for Function Wizard
	"",				// Help file (optional)
	"SabrTestSetStrikeInds.", // Function description text (optional)
	"years to expiry",				// Fn arg help text strings...
	"forward swap rate",
	"Alpha",
	"Beta",
	"Vol of vol",
	"Rho  ",
	NULL); // Arg list terminator


double __stdcall SabrTestSetStrikeInds(double Texp, double Fwd,
	double Alpha, double Beta, double VolVol, double Rho, xloper *p_is_normal)
{
#pragma EXPORT

// Check inputs:
	if(Texp <= 0.0 || Fwd <= 0.0 || Alpha <= 0.0
	|| Beta < 0.0 || Beta > 1.0 // Could relax upper limit on Beta
	|| VolVol < 0.0 // Allow zero: non-stochastic case
	|| Rho < -1.0 || Rho > 1.0)
		return -1.0;

	cpp_xloper IsNormal(p_is_normal);

	SabrTest.SetSabrParams(Alpha, Beta, VolVol, Rho, Texp);
	SabrTest.SetStrikeIndependents(Fwd);
	static double retval = 1.0; // Not thread-safe
	return retval = 1.0 - retval;
}
//============================================================

FnRegData RT_GetSabrVolOfStrike( // Instance name is not important
	"GetSabrVolOfStrike",	// Function name as in code (v11-)
	"RBBP",		// Return type and argument types (v11-)
	"",				// Function name as in code (v12+)
	"",				// Return type and argument types (v12+)
	"GetSabrVolOfStrike",	// Fn name as it appears on worksheet
	"Trigger,Strike,IsNormal",		// Argument names (Excel 11-: max 30)
	"",				// Argument names (Excel 12+: max 255)
	"Example",		// Function category for Function Wizard
	"",				// Help file (optional)
	"GetSabrVolOfStrike.", // Function description text (optional)
	"trigger",				// Fn arg help text strings...
	"strike rate",
	"(optional) True: use normal (Gaussian) model  ",
	NULL); // Arg list terminator

xloper * __stdcall GetSabrVolOfStrike(double Trigger, double Strike, xloper *pIsNormal)
{
#pragma EXPORT
	if(Strike <= 0.0)
		return p_xlErrValue;

	cpp_xloper IsNormal(pIsNormal);
	double Vol;
	SabrTest.GetVolOfStrike(Strike, Vol, IsNormal.IsTrue());
	cpp_xloper RetVal(Vol);
	return RetVal.ExtractXloper();
}