//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains functions that create cubic splines and
// interpolate them.  For licensing reasons the key passages of code are
// omitted.  (See comments in the code).  The file also contains a simple
// linear interpolation function.
//============================================================================
//============================================================================
#include <windows.h>

#include "XllAddIn.h"
#include "xl_array.h"

#define SPLINE_GRAD_START_FIXED		0x01
#define SPLINE_GRAD_END_FIXED		0x02

//======================================================================================
void make_spline(cpp_xloper &RetVal, xl4_array *input, double grad_start, double grad_end, int mode)
{
	RW rows = input->rows;
	COL cols = input->columns;

	if(cols != 2 || rows < 2)
	{
		RetVal.SetToError(xlerrValue);
		return;
	}

	double *array = input->array;
	double *x = (double *)calloc(4 * rows, sizeof(double));
	double *y = x + rows, *u = y + rows, *y2 = u + rows;
	int i, index;

// Input is expected to be in 2 columns: x-vals then y-vals
	for(index = i = 0; i < rows;)
	{
		x[i] = array[index++];
		y[i++] = array[index++];
	}

// The code here is omitted. See Numerical Recipes in C (Cambridge Press),
// Section 3.3 for the function spline(...) which can be called
// from here with suitable conversion of arguments.  The result
// of calling this function is a vector of 2nd derivatives of
// y w.r.t. x, called y2[].  This is returned to the caller
// in a cpp_xloper of type xltypeMulti as a single column.

	RetVal.InitialiseArray((RW)rows, (COL)1, y2);
	free(x); // can't free this till finished with y2 as in the same block
}
//======================================================================================
xloper * __stdcall make_spline_xl4(xl4_array *input, double grad_start, double grad_end, int mode)
{
#pragma EXPORT
	cpp_xloper RetVal;
	make_spline(RetVal, input, grad_start, grad_end, mode);
	return RetVal.ExtractXloper();
}
//======================================================================================
xloper12 * __stdcall make_spline_xl12(xl4_array *input, double grad_start, double grad_end, int mode)
{
#pragma EXPORT
	cpp_xloper RetVal;
	make_spline(RetVal, input, grad_start, grad_end, mode);
	return RetVal.ExtractXloper12();
}
//============================================================
void spline_interp(cpp_xloper &RetVal, double x, xl4_array *input, cpp_xloper &Blend)
{
// Blend is used to scale the 2nd derivatives.  This is most
// efficiently done in the last step of the cubic interpolation
// where the entire term that relates to these can be factored
// by this amount.  When 0, the effect is to produce straight
// line interpolation between points, and when 1, a normal
// cubic spline.
	double blend = (double)Blend;
	if(blend < 0.0 || blend > 1.0)
		blend = 1.0;

	RW rows = input->rows;
	COL cols = input->columns;

// Check the dimensions of the input array and then
// transpose the input (supplied as 3 columns: x-vals, y-vals
// d2y/dx2 vals) to ease access to the data in rows.
	if(cols != 3 || rows < 2 || transpose_xl4_array(input) == false)
	{
		RetVal.SetToError(xlerrValue);
		return;
	}
	double *xa = input->array;
	double *ya = input->array + rows;
	double *y2a = input->array + 2 * rows;
	double y = 0.0; // safe default given missing code below

// The code is omitted. See Numerical Recipes in C (Cambridge Press),
// Section 3.3 for the function splint(...) which can be called
// from here with suitable conversion of arguments.  The result
// of calling this function is a value y, returned to the caller
// in a cpp_xloper of type xltypeNum.
	RetVal = y;
}
//============================================================
xloper * __stdcall spline_interp_xl4(double x, xl4_array *input, xloper *pBlend)
{
#pragma EXPORT
	cpp_xloper RetVal, Blend(pBlend);
	spline_interp(RetVal, x, input, Blend);
	return RetVal.ExtractXloper();
}
//============================================================
xloper12 * __stdcall spline_interp_xl12(double x, xl4_array *input, xloper12 *pBlend)
{
#pragma EXPORT
	cpp_xloper RetVal, Blend(pBlend);
	spline_interp(RetVal, x, input, Blend);
	return RetVal.ExtractXloper12();
}
//============================================================
//============================================================
//============================================================
// Linear interpolation
//============================================================
//============================================================
//============================================================

void interp(cpp_xloper &RetVal, double x, xl4_array *yy, xl4_array *xx, bool dont_extrapolate)
{
// Check that input ranges are same size and shape and input
// is either a row or column
	if(yy->columns != xx->columns || yy->rows != xx->rows
	|| (yy->rows != 1 && yy->columns != 1))
	{
		RetVal.SetToError(xlerrValue);
		return;
	}

	int low = 0, high;

	if(yy->rows == 1)
		high = yy->columns - 1;
	else // yy->columns == 1
		high = yy->rows - 1;

	if(high == 0)
	{
		RetVal = yy->array[0];
		return;
	}

	if(x < xx->array[0] || x > xx->array[high])
	{
		if(dont_extrapolate)
		{
			RetVal.SetToError(xlerrNum);
			return;
		}

		if(x < xx->array[0])
			high = 1;
		else
			low = high - 1;
	}
	else
	{
		int i;
		while(high - low > 1)
		{
			if(xx->array[i = (high + low) >> 1] > x)
				high = i;
			else
				low = i;
		}
	}
	RetVal = yy->array[low] + (x - xx->array[low]) *
		(yy->array[high] - yy->array[low]) / (xx->array[high] - xx->array[low]);
}
//=================================================================================
xloper * __stdcall interp_xl4(double x, xl4_array *yy, xl4_array *xx, xloper *p_dont_extrap)
{
#pragma EXPORT
	cpp_xloper RetVal(p_dont_extrap); // temporary use of this
	interp(RetVal, x, yy, xx, RetVal.IsTrue());
	return RetVal.ExtractXloper();
}
//=================================================================================
xloper12 * __stdcall interp_xl12(double x, xl4_array *yy, xl4_array *xx, xloper12 *p_dont_extrap)
{
#pragma EXPORT
	cpp_xloper RetVal(p_dont_extrap); // temporary use of this
	interp(RetVal, x, yy, xx, RetVal.IsTrue());
	return RetVal.ExtractXloper12();
}
//============================================================
//============================================================
//============================================================
// Bilinear interpolation
//============================================================
//============================================================
//============================================================
bool get_bracket(int n_vals, double *values, double search_val, int &low_index, int &high_index)
{
	if(search_val < values[0])
	{
		low_index = -1; // out of bounds
		high_index = 0;
		return true;
	}

	if(search_val > values[n_vals - 1])
	{
		low_index = n_vals - 1;
		high_index = n_vals; // out of bounds
		return true;
	}

// Use bisection search to find the bracketing rows for search_val
	int i;
	high_index = n_vals;
	low_index = 0;

	while(high_index - low_index > 1)
	{
// max(i) == n_vals - 1, min(i) == 0
		i = (high_index + low_index) >> 1;

		if(values[i] > search_val)
			high_index = i;
		else
			low_index = i;
	}
	return true;
}
//============================================================
// Calculate the value t = (x - x[j]) / (x[j+1] - x[j]) where
// (x[j], x[j+1]) brackets x. Also return the indices low_index
// and high_index
//
// Used in bilinear interpolation:
// y(x1, x2) = (1-t1)(1-t2).yll + t1(1-t2).yhl + t1.t2.yhh + (1-t1)t2.ylh
//============================================================
bool bilinear_calc_t(int n_vals, double *x_vals, double x, int &low_index, int &high_index, double &t)
{
	if(!get_bracket(n_vals, x_vals, x, low_index, high_index))
		return false;

	if(low_index == -1)			low_index = 0;
	if(high_index == n_vals)	high_index--;

	if(x_vals[high_index] == x_vals[low_index])
// Could be a problem with the x's or high_index == low_index
		t = 0.0;
	else
		t = (x - x_vals[low_index])/(x_vals[high_index]-x_vals[low_index]);
	return true;
}
//============================================================
bool bilinear_interp(double &y, double row_x, double *row_x_vals,
	double col_x, double *col_x_vals, double *y_vals,
	int num_rows, int num_cols, bool check_x_vals)
{
// If check_x_vals == true, check the x-vals are all ascending
	if(check_x_vals)
	{
		double d, last = row_x_vals[0];

		for(int i = 1; i < num_rows; i++, last = d)
			if(last >= (d = row_x_vals[i]))
				return false;

		last = col_x_vals[0];

		for(i = 1; i < num_cols; i++, last = d)
			if(last >= (d = col_x_vals[i]))
				return false;
	}

// Check that row_x and col_x are within the repsective ranges
	if(row_x < row_x_vals[0]
	|| row_x > row_x_vals[num_rows - 1]
	|| col_x < col_x_vals[0]
	|| col_x > col_x_vals[num_cols - 1])
		return false;

// Find the x-pair in the column of x-values that correspond to
// each row (row_x_vals) that bracket the given value of x, and
// calculate the parameter t that determines how far from the
// first of the two bracketting points x is.
	int row_low_index, row_high_index;
	double row_t;

	if(!bilinear_calc_t(num_rows, row_x_vals,
		row_x, row_low_index, row_high_index, row_t))
		return false;

// Find the x-pair in the row of x-values that correspond to each
// column (col_x_vals) that bracket the given value of x, and
// calculate the parameter t that determines how far from the
// first of the two bracketting points x is.
	int col_low_index, col_high_index;
	double col_t;

	if(!bilinear_calc_t(num_cols, col_x_vals,
		col_x, col_low_index, col_high_index, col_t))
		return false;

// Extract the four y-values corresponding to the corners of the
// rectangle described by row_low_index, row_high_index,
// col_low_index, col_high_index.
//			col_index
//			low		high
//  y_rl_cl---------------y_rl_ch
//	 |						|	low
//	 |						|
//	 |						|		row_index
//	 |						|
//	 |						|	high
//  y_rh_cl---------------y_rh_ch

	double y_rl_cl = y_vals[num_cols * row_low_index + col_low_index];
	double y_rh_cl = y_vals[num_cols * row_high_index + col_low_index];
	double y_rl_ch = y_vals[num_cols * row_low_index + col_high_index];
	double y_rh_ch = y_vals[num_cols * row_high_index + col_high_index];

// Perform the bilinear interpolation
	double ll = (1-row_t) * (1-col_t);
	double lh = (1-row_t) * col_t;
	double hh = row_t * col_t;
	double hl = row_t * (1-col_t);

	y = ll * y_rl_cl + lh * y_rl_ch + hh * y_rh_ch + hl * y_rh_cl;
	return true;
}
//============================================================


FnRegData RT_bilinear_interp_xl4( // Instance name is not important
	"bilinear_interp_xl4",	// Function name as in code (v11-)
	"RBKBKKP",		// Return type and argument types (v11-)
	"",				// Function name as in code (v12+)
	"",				// Return type and argument types (v12+)
	"BilinearInterp",	// Fn name as it appears on worksheet
	"rowX,rowXvals,colX,colXvals,Yvals,CheckXvals",		// Argument names (Excel 11-: max 30)
	"",				// Argument names (Excel 12+: max 255)
	"Example",		// Function category for Function Wizard
	"",				// Help file (optional)
	"Bilinear interpolation.", // Function description text (optional)
	"row x value",				// Fn arg help text strings...
	"column of row x values in ascending order",
	"column x value",
	"row of columns x values in ascending order",
	"array of y values",
	"(optional) True: force checking of input x values  ",
	NULL); // Arg list terminator

xloper * __stdcall bilinear_interp_xl4(double row_x, xl4_array *row_x_vals,
	double col_x, xl4_array *col_x_vals, xl4_array *y_vals, xloper *pCheckXvals)
{
// Check the input array sizes
	if(y_vals->columns != col_x_vals->columns
	|| y_vals->rows != row_x_vals->rows
	|| col_x_vals->rows != 1 || row_x_vals->columns != 1)
		return p_xlErrNum;

	cpp_xloper Op(pCheckXvals);
	double y;

	if(!bilinear_interp(y, row_x, row_x_vals->array,
		col_x, col_x_vals->array, y_vals->array,
		y_vals->rows, y_vals->columns, Op.IsTrue()))
		return p_xlErrNum;

	Op = y;
	return Op.ExtractXloper();
}
