//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains functions that calculate the cumulative normal
// distribution and its inverse and that calculate random samples from this
// distribution using the Box-Muller transformation.
//============================================================================
//============================================================================
#include <windows.h>
#include <math.h>

#include "XllAddIn.h"
#include "Stack.h"

//=======================================================================
// Algorthm used by Excel 2003, wrapped in a thread-safe structure to
// make it safe to call from multiple threads.

// Wichman, B.A. and I.D. Hill, Algorithm AS 183:
// An Efficient and Portable Pseudo-Random Number Generator,
// Applied Statistics, 31, 188-190, 1982.

// Wichman, B.A. and I.D. Hill
// Building a Random-Number Generator, BYTE, pp. 127-128, March 1987.

#define SEED_1_DFT 8000
#define SEED_2_DFT 16000
#define SEED_3_DFT 24000

struct ts_rand
{
	ts_rand(int ix_seed, int iy_seed, int iz_seed)
	{
		if(gExcelVersion12plus)
			InitializeCriticalSection(&cs_rand);
		set_seeds(ix_seed, iy_seed, iz_seed);
	}

	~ts_rand(void)
	{
		if(gExcelVersion12plus)
			DeleteCriticalSection(&cs_rand);
	}

	bool set_seeds(int ix_seed, int iy_seed, int iz_seed)
	{
		if(ix_seed < 1 || ix_seed >= 30000) ix_seed = SEED_1_DFT;
		if(iy_seed < 1 || iy_seed >= 30000) iy_seed = SEED_2_DFT;
		if(iz_seed < 1 || iz_seed >= 30000) iz_seed = SEED_3_DFT;
		if(gExcelVersion12plus)
			EnterCriticalSection(&cs_rand);
		ix = ix_seed;
		iy = iy_seed;
		iz = iz_seed;
		if(gExcelVersion12plus)
			LeaveCriticalSection(&cs_rand);
		return true;
	}
	double get(void)
	{
		double d, x, y, z;
		if(gExcelVersion12plus)
			EnterCriticalSection(&cs_rand);
		x = ix = (171 * ix) % 30269;
		y = iy = (172 * iy) % 30307;
		z = iz = (170 * iz) % 30323;
		if(gExcelVersion12plus)
			LeaveCriticalSection(&cs_rand);
		d = x / 30269.0 + y / 30307.0 + z / 30323.0;
		return modf(d, &d); // &d passed but integer part not used
	}
private:
	CRITICAL_SECTION cs_rand; // Only used when version is 12+
	int ix, iy, iz;
};

ts_rand rand_xl2003(0, 0, 0); // Use default seeds


#define PI	3.14159265358979323846264
//==============================================================
void generate_BM_pair(double &z1, double &z2)
{
// Use Excel 2003's algorithm to generate std random numbers.
// More reliable than earlier Excel algorithms and calling
// this implementation of it is much faster than calling
// back into Excel with xlfRand.
	double r1 = rand_xl2003.get();
	double r2 = rand_xl2003.get();
// Use half-angle tan formulae to minimise trig fn calls
	double t = tan(r2 * PI), tt = t * t;
	r1 = sqrt(-2.0 * log(r1)) / (1.0 + tt);
	z1 = r1 * (1.0 - tt);
	z2 = r1 * 2.0 * t;
}
//==============================================================
void __stdcall nsample_BM_pair(xl4_array *p_array)
{
#pragma EXPORT

	size_t array_size = p_array->columns * p_array->rows;

	if(array_size == 2)
		generate_BM_pair(p_array->array[0], p_array->array[1]);
	else
		memset(p_array->array, 0, array_size * sizeof(double));
}
//==============================================================

// Define this to be greater than the maximum number of threads that
// could reasonably be expected to run, so that pushed values are not
// lost.
#define NSAMPLE_BM_STACK_SIZE	100
simple_stack nsample_BM_stack(NSAMPLE_BM_STACK_SIZE);

//==============================================================
// Need to use a thread-safe stack to store and retrieve
// values as many threads could be accessing (reading/storing)
// values simultaneously
double __stdcall nsample_BM(void)
{
#pragma EXPORT
	double z1, z2;
	if(nsample_BM_stack.pop(z2))
		return z2;
	generate_BM_pair(z1, z2);
	nsample_BM_stack.push(z2); // save for next call
	return z1;
}
//=======================================================================
double __stdcall xll_rand_xl4(xloper *pSeed1, xloper *pSeed2, xloper *pSeed3)
{
#pragma EXPORT
	if(pSeed1->xltype == xltypeNum
	&& pSeed2->xltype == xltypeNum
	&& pSeed3->xltype == xltypeNum)
	{
		rand_xl2003.set_seeds((int)pSeed1->val.num,
			(int)pSeed2->val.num, (int)pSeed3->val.num);
	}
	return rand_xl2003.get();
}
//=======================================================================
double __stdcall xll_rand_non_vol_xl4(double trigger, xloper *pSeed1, xloper *pSeed2, xloper *pSeed3)
{
#pragma EXPORT
	return xll_rand_xl4(pSeed1, pSeed2, pSeed3);
}
//=======================================================================
double __stdcall xll_rand_xl12(xloper12 *pSeed1, xloper12 *pSeed2, xloper12 *pSeed3)
{
#pragma EXPORT
	if(pSeed1->xltype == xltypeNum
	&& pSeed2->xltype == xltypeNum
	&& pSeed3->xltype == xltypeNum)
	{
		rand_xl2003.set_seeds((int)pSeed1->val.num,
			(int)pSeed2->val.num, (int)pSeed3->val.num);
	}
	return rand_xl2003.get();
}
//=======================================================================
double __stdcall xll_rand_non_vol_xl12(double trigger, xloper12 *pSeed1, xloper12 *pSeed2, xloper12 *pSeed3)
{
#pragma EXPORT
	return xll_rand_xl12(pSeed1, pSeed2, pSeed3);
}
//=======================================================================
// Numerical approximation to the cumulative normal dsitribution
//=======================================================================
// double cndist(double d)
//
// See Abramowitz & Stegun 26.2.17  |error| < 7.5e-8 
//=======================================================================

#define B1					0.31938153
#define B2					-0.356563782
#define B3					1.781477937
#define B4					-1.821255978
#define B5					1.330274429
#define PP					0.2316419
#define ROOT_2PI			2.506628274631
#define NEG_LN_ROOT_2PI		-0.918938533204673
//=======================================================================
double cndist(double d)
{
	if(d == 0.0) return 0.5;
	double t = 1.0 / (1.0 + PP * fabs(d));
	double e = exp(NEG_LN_ROOT_2PI - 0.5 * d * d);
	double n = ((((B5 * t + B4) * t + B3) * t + B2) * t + B1) * t;
	return (d > 0.0) ? 1.0 - e * n : e * n;
}
//=======================================================================
#define MAX_CNDIST_ITERS	500
//
// Calculates cumulative normal distribution to highest accuracy
// supported using Taylor series subject to cummulative rounding
// errors.
//
//=======================================================================
double cndist_taylor(double d, int &iterations)
{
	if(fabs(d) > 5.5)
	{
// Small difference between the cndist() approximation and the real
// thing in the tails, although this might upset some pdf functions,
// where kinks in the gradient create large jumps in the pdf. Should
// ideally be replaced with a more accurate method for the tails.
		iterations = 0;
		return cndist(d);
	}
	double d2 = d * d, last_sum = 0, sum = 1.0, factor = 1.0, k2;
	for(int k = 1; k <= MAX_CNDIST_ITERS; k++)
	{
		k2 = k << 1;
		sum += (factor *= d2 * (1.0 - k2) / k2 / (k2 + 1.0));

		if(last_sum == sum)
			break;

		last_sum = sum;
	}
	iterations = k;
	return 0.5 + sum * d / ROOT_2PI;
}
//=====================================================================
// Excel add-in interface functions
//=====================================================================
double __stdcall ndist(double d)
{
#pragma EXPORT

// Excel's own functions are not reliable for versions 9 or lower.
// If unititalised, gExcelVersion is zero, so will use the DLL
// code to be safe.
	if(gExcelVersion <= 9)
	{
		int iterations; // not used here
		return cndist_taylor(d, iterations);
	}
// Else use Excel's own NORMSDIST function
	cpp_xloper Op(d);
	Op.Excel(xlfNormsdist, 1, &Op); // re-use Op
	return (double)Op;
}
//=====================================================================
#define NDINV_ITER_LIMIT		50
#define NDINV_EPSILON			1e-12 // How precise do we want to be
#define NDINV_FIRST_NUDGE		1e-7

// Minimum change in answer from one iteration to the next
#define NDINV_DELTA				1e-10

// Approximate working limits of Excel's NORMSINV() function in Excel 2000
#define NORMSINV_LOWER_LIMIT	3.024e-7
#define NORMSINV_UPPER_LIMIT	0.999999

//=====================================================================
void ndist_inv(double prob, cpp_xloper &RetVal)
{
	if(gExcelVersion > 9) // OK to use Excel's own NORMSINV function
	{
		RetVal = prob;
		RetVal.Excel(xlfNormsinv, 1, &RetVal);
		return;
	}

// For versions below 9 (Excel 2000) Excel's own functions are not reliable
// so use a Taylor series
	if(prob <= 0.0 || prob >= 1.0)
	{
		RetVal.SetToError(xlerrNum);
		return;
	}

// Get a (pretty) good first approximation using Excel's NORMSINV()
// worksheet function. First check that prob is within NORMSINV's
// working limits
	int iterations;
	double v1, v2, p1, p2, pdiff, temp;

	if(prob < NORMSINV_LOWER_LIMIT)
	{
		v2 = (v1 = -5.0) - NDINV_FIRST_NUDGE;
	}
	else if(prob > NORMSINV_UPPER_LIMIT)
	{
		v2 = (v1 = 5.0) + NDINV_FIRST_NUDGE;
	}
	else
	{
		RetVal = prob;
		RetVal.Excel(xlfNormsinv, 1, &RetVal);
		if(!RetVal.IsNum()) // shouldn't ever be true
		{
			RetVal.SetToError(xlerrNum);
			return;
		}
		v2 = (double)RetVal;
		v1 = v2 - NDINV_FIRST_NUDGE;
	}

// Use a secant method to make the result consistent with the
// cndist_taylor() function
	p2 = cndist_taylor(v2, iterations) - prob;

	if(fabs(p2) <= NDINV_EPSILON)
	{
		RetVal = v2;
		return; // already close enough
	}

	p1 = cndist_taylor(v1, iterations) - prob;

	for(short i = NDINV_ITER_LIMIT; --i;)
	{
		if(fabs(p1) <= NDINV_EPSILON || (pdiff = p2 - p1) == 0.0)
		{
// Result is close enough, or need to avoid divide by zero
			RetVal = v1;
			return;
		}
		temp = v1;
		v1 = (v1 * p2 - v2 * p1) / pdiff;

		if(fabs(v1 - temp) <= NDINV_DELTA) // not much improvement
		{
			RetVal = v1;
			return;
		}
		v2 = temp;
		p2 = p1;
		p1 = cndist_taylor(v1, iterations) - prob;
	}
	RetVal.SetToError(xlerrValue);
}
