//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains functions that relate to the pricing of constant
// maturity swap (CMS) derivatives.
//============================================================================
//============================================================================
#include <windows.h>
#include <math.h>

#include "XllAddIn.h"
#include "SABR.h"
#include "Black.h"


//=================================================================
// Functions in this example code are simplified so that, for example,
// currency and swap payment basis, cash-settlement conventions, etc. are
// not specified. Swap rates are assumed to be bond basis annuallly paid.
// For real-world use, these assumptions would need to be replaced.
//=================================================================
// Returns the IRR unit PV of a vanilla swap at a given rate.
// Equvalent to Excel's PV() function.  Assumption here and in Excel
// function is that the rate is Act/Act = 30/360 unadjusted.
//=================================================================
double swap_unit_pv(int term_yrs, int fixed_pmts_pa, double rate)
{
	double D = 1.0 / (1.0 + rate / fixed_pmts_pa);
	return (1.0 - pow(D, fixed_pmts_pa * term_yrs)) / rate * fixed_pmts_pa;
}
//=================================================================
// Calculates the cost of a replicating portfolio for a CMS delta_T
// caplet using a strike step of delta_S.  Returns 0 if the process
// failed for any reason.

// Returns the cost of the portfolio as valued at the start date of
// the underlying swap.
//=================================================================
#define MAX_ITERATIONS	10000 // Just to stop the loop running away

// *********** NOTE ***********
// Code is changed from the listing in the book: the argument
// SabrWrapper &Sabr is added and initialisation of this is now
// assumed to be taken care of by the caller.
// Check on input max_strike now only done if is_call is true.

double CMS_portfolio_cost(double Texp, double delta_T, double fwd_swap,
						double short_rate, double strike, int term_yrs,
						int fixed_pmts_pa, bool is_call, double delta_S,
						double max_strike, double min_value_increment,
						double lambda, SabrWrapper &Sabr)
{
// Check the inputs
	if(Texp <= 0.0 || delta_T <= 0.0 || fwd_swap <= 0.0
	|| short_rate <= 0.0 || strike <= 0.0 || term_yrs <= 0
	|| (fixed_pmts_pa != 1 && fixed_pmts_pa != 2 && fixed_pmts_pa != 4 && fixed_pmts_pa != 12)
	|| delta_S <= 0.0 || (is_call && max_strike <= 0.0) || min_value_increment <= 0.0)
		return 0.0;

	if(!is_call) // for floorlet, add -ve increment
		delta_S = -delta_S;

// Initialise the strike-independant quantities to speed
// up calculation of vols for different strikes of what
// is always the same option.
	Sabr.SetStrikeIndependents(fwd_swap);

// Create an instance of BlackOption class for vanilla swaption
// pricing.  For now, just set up the things that don't change.
	BlackOption Black;
	Black.SetTexp(Texp);
	Black.SetFwd(fwd_swap);

	double P_fwd; // the unit PV of the swap at the current forward rate
	double P; // the unit PV of the swap at the fixing rate
	double N; // the swaption notional of the strike being added
	double last_X = 0.0, X = 0.0, sum_N = 0.0, sum_iN = 0.0;
	double i_plus_lambda;
	double black_vol, black_price, last_black_price = MAX_DOUBLE;
	double inv_delta_T = 1.0 / delta_T;

// Assume that initial swap and short rates are same ratio and
// use this to calculate short_r given a swap fixing rate
	double current_ratio = short_rate / fwd_swap;

// Set initial fixing rate
	double fixing = strike + delta_S * lambda;
	P_fwd = swap_unit_pv(term_yrs, fixed_pmts_pa, fwd_swap);

	for(int i = 0; i < MAX_ITERATIONS; i++)
	{
// Calculate the unit PV of a swap at this fixing rate, at
// which the value of the portfolio is about to be recalculated.
		P = swap_unit_pv(term_yrs, fixed_pmts_pa, fixing);

// Use very simplified assumption for the short rate given this swap fixing
		short_rate = current_ratio * fixing;

// Calculate the notional amount of payer swaption
// at strike = (fixing - lambda * delta_S)
		i_plus_lambda = i + lambda;
		N = (i_plus_lambda / (inv_delta_T + short_rate) / P
			+ sum_iN - i_plus_lambda * sum_N) / lambda;

// Calculate the cost of the vanilla swaption at this strike
		if(!Sabr.GetVolOfStrike(strike, black_vol, false)) // false: get Black vol
			return 0.0; // Couldn't get a good vol

		Black.SetStrike(strike);
		Black.SetVol(black_vol);
		Black.Calc(false); // false: don't calculate greeks
		black_price = (is_call ? Black.GetCallPrice() : Black.GetPutPrice());

// Check if more out-of-the-money option is more expensive than
// the last option.  This should, in theory, not happen, but
// is possible using the SABR expressions, where the limits of
// the underlying assumptions have been exceeded.  If so, just
// terminate the building of the portfolio.
		if(last_black_price <= black_price)
			break;

// Calculate and add the cost of this notional of this strike to the portfolio
		X += black_price * P_fwd * N;

// Accumulate portfolio hedge sensitivities here (omitted)

// Check if the change in value is less than the specified limit
		if(fabs(last_X - X) < min_value_increment)
			break;

// Increment the sums, strike and fixing for the next loop
		strike += delta_S;
		if(is_call && strike > max_strike)
			break;

		if((fixing += delta_S) <= 0.0)
			break;

		sum_N += N;
		sum_iN += i * N;
		last_X = X;
		last_black_price = black_price;
	}
	return X;
}
//=================================================================
// XLL interface function to CMS_portfolio_cost()
//=================================================================

// *********** NOTE ***********
// Code is changed from the listing in the book: the SABR parameters
// Alpha, Beta, Rho and VolVol are added and initialisation of the
// SabrWrapper Sabr is done and a reference of this passed to
// the core function CMS_portfolio_cost.

double __stdcall CmsPortfolioCost(double Texp, double delta_T,
	double fwd_swap, double short_rate, double strike, int term_yrs,
	int fixed_pmts_pa, int is_call, double delta_S, double max_strike,
	double min_value_increment, double lambda,
	double Alpha, double Beta, double Rho, double VolVol)
{
#pragma EXPORT
// Initialise the wrapper to the SABR functions and set safe defaults
// if inputs out of range
	SabrWrapper Sabr(Alpha, Beta, VolVol, Rho, Texp);

// Inputs are checked are done in CMS_portfolio_cost(), so don't bother here
	return CMS_portfolio_cost(Texp, delta_T, fwd_swap,
		short_rate, strike, term_yrs, fixed_pmts_pa,
		is_call == 0 ? false : true,
		delta_S, max_strike, min_value_increment, lambda, Sabr);
}
