//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains examples that extend the functionality of
// functions such as MATCH, COUNTIF and SUMIF to accommodate more multiple
// criteria.
//============================================================================
//============================================================================
#include <windows.h>
#include "XllAddIn.h"

#define NUM_MULTI_FN_RANGES		5

#define DO_MULTI_MATCH		1
#define DO_MULTI_SUM		2
#define DO_MULTI_COUNT		4
#define DO_MULTI_AVERAGE	8

//=================================================================
void do_multi(cpp_xloper &RetVal, cpp_xloper &SumRange, cpp_xloper *Args, int operation)
{
	if((operation & (DO_MULTI_SUM | DO_MULTI_AVERAGE))
	&& !SumRange.IsType(xltypeMulti))
	{
		RetVal.SetToError(xlerrValue);
		return;
	}

// Find the last non-missing value/range pair
	int num_searches = 0;
	int i, index = 0;

	do
	{
		if(Args[index++].IsType(xltypeMissing | xltypeNil | xltypeErr)
		|| !Args[index++].IsType(xltypeMulti))
			break;
	}
	while(++num_searches < NUM_MULTI_FN_RANGES);

	if(!num_searches)
	{
		RetVal.SetToError(xlerrValue);
		return;
	}

// Check that all the input arrays are the same shape and size
	RW rows, temp_rows;
	COL columns, temp_columns;

	if(operation & (DO_MULTI_SUM | DO_MULTI_AVERAGE))
		SumRange.GetArraySize(rows, columns);
	else
		Args[0].GetArraySize(rows, columns);

// Check that input is either single row or single column
	if(rows != 1 && columns != 1)
	{
		RetVal.SetToError(xlerrValue);
		return;
	}

	for(index = 2 * num_searches; --index; )
	{
		Args[index].GetArraySize(temp_rows, temp_columns);
		if(rows != temp_rows || columns != temp_columns)
		{
			RetVal.SetToError(xlerrValue);
			return;
		}
	}

// Simple search does not assume search ranges are sorted and
// looks for an exact match
	DWORD limit = rows * columns, offset;
	double temp, sum = 0.0;
	int count = 0;
	for(offset = 0; offset < limit; offset++)
	{
		for(index = i = 0; i < num_searches; i++, index += 2)
			if(!Args[index + 1].ArrayEltEq(offset, Args[index]))
				break;

		if(i == num_searches) // Match found!
		{
			if(operation & (DO_MULTI_SUM | DO_MULTI_AVERAGE))
			{
				if(SumRange.GetArrayElt(offset, temp))
					sum += temp;

				if(operation == DO_MULTI_AVERAGE)
					count++;
			}				
			else if(operation == DO_MULTI_COUNT)
			{
				count++;
			}
			else if(operation == DO_MULTI_MATCH)
			{
		// Increment the offset as INDEX() counts from 1
				RetVal = (double)(offset + 1);
				return;
			}
		}
	}
	if(operation == DO_MULTI_SUM)
		RetVal = sum;
	else if(operation == DO_MULTI_COUNT)
		RetVal = (double)count;
	else if(operation == DO_MULTI_AVERAGE)
		RetVal = sum / count;
	else if(operation == DO_MULTI_MATCH)
		RetVal.SetToError(xlerrNA);
}
//=================================================================
xloper * __stdcall match_multi_xl4(
		xloper *value1, xloper *range1,
		xloper *value2, xloper *range2,
		xloper *value3, xloper *range3,
		xloper *value4, xloper *range4,
		xloper *value5, xloper *range5)
{
#pragma EXPORT

// Arguments are registered as type P so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal, SumRange; // SumRange not used
	do_multi(RetVal, SumRange, Args, DO_MULTI_MATCH);
	return RetVal.ExtractXloper();
}
//=================================================================
xloper12 * __stdcall match_multi_xl12(
		xloper12 *value1, xloper12 *range1,
		xloper12 *value2, xloper12 *range2,
		xloper12 *value3, xloper12 *range3,
		xloper12 *value4, xloper12 *range4,
		xloper12 *value5, xloper12 *range5)
{
#pragma EXPORT

// Arguments are registered as type Q so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal, SumRange; // SumRange not used
	do_multi(RetVal, SumRange, Args, DO_MULTI_MATCH);
	return RetVal.ExtractXloper12();
}
//=================================================================
xloper * __stdcall sum_if_multi_xl4(xloper *sum_range,
		xloper *value1, xloper *range1,
		xloper *value2, xloper *range2,
		xloper *value3, xloper *range3,
		xloper *value4, xloper *range4,
		xloper *value5, xloper *range5)
{
#pragma EXPORT

// Arguments are registered as type P so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal;
	cpp_xloper SumRange(sum_range);
	do_multi(RetVal, SumRange, Args, DO_MULTI_SUM);
	return RetVal.ExtractXloper();
}
//=================================================================
xloper12 * __stdcall sum_if_multi_xl12(xloper *sum_range,
		xloper12 *value1, xloper12 *range1,
		xloper12 *value2, xloper12 *range2,
		xloper12 *value3, xloper12 *range3,
		xloper12 *value4, xloper12 *range4,
		xloper12 *value5, xloper12 *range5)
{
#pragma EXPORT

// Arguments are registered as type Q so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal;
	cpp_xloper SumRange(sum_range);
	do_multi(RetVal, SumRange, Args, DO_MULTI_SUM);
	return RetVal.ExtractXloper12();
}
//=================================================================
xloper * __stdcall count_if_multi_xl4(
		xloper *value1, xloper *range1,
		xloper *value2, xloper *range2,
		xloper *value3, xloper *range3,
		xloper *value4, xloper *range4,
		xloper *value5, xloper *range5)
{
#pragma EXPORT

// Arguments are registered as type P so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal, SumRange; // SumRange not used
	do_multi(RetVal, SumRange, Args, DO_MULTI_COUNT);
	return RetVal.ExtractXloper();
}
//=================================================================
xloper12 * __stdcall count_if_multi_xl12(
		xloper12 *value1, xloper12 *range1,
		xloper12 *value2, xloper12 *range2,
		xloper12 *value3, xloper12 *range3,
		xloper12 *value4, xloper12 *range4,
		xloper12 *value5, xloper12 *range5)
{
#pragma EXPORT

// Arguments are registered as type Q so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal, SumRange; // SumRange not used
	do_multi(RetVal, SumRange, Args, DO_MULTI_COUNT);
	return RetVal.ExtractXloper12();
}
//=================================================================
xloper * __stdcall average_if_multi_xl4(xloper *average_range,
		xloper *value1, xloper *range1,
		xloper *value2, xloper *range2,
		xloper *value3, xloper *range3,
		xloper *value4, xloper *range4,
		xloper *value5, xloper *range5)
{
#pragma EXPORT

// Arguments are registered as type P so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal;
	cpp_xloper AverageRange(average_range);
	do_multi(RetVal, AverageRange, Args, DO_MULTI_AVERAGE);
	return RetVal.ExtractXloper();
}
//=================================================================
xloper12 * __stdcall average_if_multi_xl12(xloper *average_range,
		xloper12 *value1, xloper12 *range1,
		xloper12 *value2, xloper12 *range2,
		xloper12 *value3, xloper12 *range3,
		xloper12 *value4, xloper12 *range4,
		xloper12 *value5, xloper12 *range5)
{
#pragma EXPORT

// Arguments are registered as type Q so range references are
// already converted to xltypeMulti.
	cpp_xloper Args[2 * NUM_MULTI_FN_RANGES] =
	{
		value1, range1, value2, range2, value3, range3,
		value4, range4, value5, range5,
	};
	cpp_xloper RetVal;
	cpp_xloper AverageRange(average_range);
	do_multi(RetVal, AverageRange, Args, DO_MULTI_AVERAGE);
	return RetVal.ExtractXloper12();
}