//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
#pragma once

// Returns the Black vol (if is_normal is false) of a vanilla swaption.
bool sabr_vol(double FwdRate, double Strike, double Texp,
			  double Alpha, double Beta, double VolVol,
			  double Rho, double &Vol, bool is_normal);

// Simple class used in CMS pricing to hide implementation of
// SABR model and provide simple means of getting volatility
// for a given option and strike.  Provides two methods to
// further speed up the calculation of a sequence of 
// volatilities for the same option but of different strikes

#define MIN_VOL	(0.00001)

class SabrWrapper
{
public:
	SabrWrapper(void) {}
	SabrWrapper(double Alpha, double Beta, double VolVol, double Rho, double Texp)
		: m_Alpha(Alpha), m_Beta(Beta), m_VolVol(VolVol), m_Rho(Rho), m_Texp(Texp)
	{
		CheckParams();
	}

	void SetSabrParams(double Alpha, double Beta, double VolVol, double Rho, double Texp)
	{
		m_Alpha = Alpha;
		m_Beta = Beta;
		m_VolVol = VolVol;
		m_Rho = Rho;
		m_Texp = Texp;
		CheckParams();
	}

	bool SetStrikeIndependents(double fwd_rate);
	bool GetVolOfStrike(double Strike, double &Vol, bool is_normal);

	void CheckParams(void)
	{
		if(m_Alpha <= 0.0)
			m_Alpha = MIN_VOL;

		if(m_Beta < 0.0)
			m_Beta = 0.0;
		else if(m_Beta > 2.0) // Allow values of Beta > 1
			m_Beta = 1.0;

		if(m_VolVol < 0.0)
			m_VolVol = MIN_VOL;

		if(fabs(m_Rho) > 1.0)
			m_Rho = 0.0;

		if(m_Texp < 0.0)
			m_Texp = 0.0;
	}

protected:
	double m_Texp;
	double m_Alpha;
	double m_Beta;
	double m_VolVol;
	double m_Rho;

// Strike-independent values:
	double m_FwdRate; // fwd rate
	double m_lambda_B; // = (m_Beta - 1)^2
	double m_lambda_N; // = m_Beta * (m_Beta - 2) = m_lambda_B - 1
	double m_a_B; // = m_lambda_B / 24
	double m_a_N; // = m_lambda_N / 24
	double m_b; // = m_Rho * m_Beta * m_VolVol / 4
	double m_c; // = (2 - 3 * m_Rho^2)/24 * m_VolVol^2
	double m_d; // = (m_Beta - 1) / 2
};
