//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================

//============================================================================
//============================================================================
// This source file (and header file) contain the definition of a class for
// managing and accessing Excel names, both worksheet names and DLL-internal
// names.  It contains examples of the use of this class and an STL container
// for managing DLL-internal Excel names.
//============================================================================
//============================================================================
#include <windows.h>
#include <stdio.h>
#include <time.h>

// Needed because of Microsoft bug: See Article ID: Q167355 
#pragma warning(disable:4786)

#include "XllAddIn.h"
#include "XllNames.h"

#include <map>
#include <string>

//==================================================
//==================================================
//==================================================
//
// Structures and functions for management of list
// of internal xll Excel names.
//
//==================================================
//==================================================
//==================================================

// Local class. Only requires one specialized contructor
// Default constructor is hidden.
// No copy/assignment constructors required

struct name_list_elt
{
	name_list_elt(xlName &R, void *fn_ptr)
		: m_R(R), m_fn_ptr(fn_ptr) {}

	xlName m_R;
	void * m_fn_ptr;

private:
	name_list_elt(); // no implementation
};

// Internal range name map will hold a copy of the name in
// a std::string which we'll map to name_list_elt

typedef std::map<std::string, name_list_elt*> MAP_NAMES;

MAP_NAMES mapNames;

char *make_unique_name(void);
int __stdcall clean_xll_name_list(void);
int __stdcall delete_all_xll_names(void);

//==================================================
bool add_name_record(void *fn_ptr, xlName &Range)
{
	const char *range_name = Range.GetName();

	if(!range_name)
		return false;	// can't add when there's no name
	
	std::string sRangeName = range_name;
	
	free((void *)range_name);	// release deep copy from GetName

	name_list_elt* pElement = new name_list_elt(Range, fn_ptr);
	
	if(!pElement)
		return false;

// Not strictly necesary but clear out any rubbish before
// looking for existing name.  Could be inefficient if
// list of internal names is long, and could be safely
// removed.
	clean_xll_name_list();

	MAP_NAMES::iterator iterNames = mapNames.find(sRangeName);

	if(iterNames != mapNames.end())
		return false; // Range Name already in map	

	xlName::IncrNumInternalNames();
	mapNames[sRangeName] = pElement;
	return true;
}
//==================================================
// Return the funtion pointer that was associated
// with this DLL-internal name
//==================================================
void *find_name_record_fn_ptr(xlName &Range)
{
	const char *range_name = Range.GetName();

	if(!range_name)
		return false;	// can't add when there's no name
	
	std::string sRangeName = range_name;
	
	free((void *)range_name);	// release deep copy from GetName

// Not strictly necesary but clear out any rubbish before
// looking for existing name.  Could be inefficient if
// list of internal names is long, and could be safely
// removed.
	clean_xll_name_list();

	MAP_NAMES::iterator iterNames = mapNames.find(sRangeName);

	if(iterNames == mapNames.end())
		return false; // Range Name not found
	
	name_list_elt* pElement = (*iterNames).second;

	return pElement->m_fn_ptr;
}
//==================================================
int __stdcall clean_xll_name_list(void)
{
#pragma EXPORT

	if(mapNames.size() == 0)
		return 0;

	MAP_NAMES::iterator iterNames = mapNames.begin();

	while(iterNames != mapNames.end())
	{
		name_list_elt* pElement = (*iterNames).second;		
	
		if(!pElement->m_R.RefreshRef())
		{
// If the range is no longer valid remove it
			pElement->m_R.Delete();
			delete pElement;

// erase returns iterator to next valid entry or to end()
			iterNames = mapNames.erase(iterNames);
			xlName::DecrNumInternalNames();
			continue;
		}
		iterNames++;
	}
	return 1;
}
//==================================================
int __stdcall delete_all_xll_names(void)
{
#pragma EXPORT
	if(mapNames.size() == 0)
		return 0;

	MAP_NAMES::iterator iterNames = mapNames.begin();

	while(iterNames != mapNames.end())
	{
		name_list_elt* pElement = (*iterNames).second;		
		pElement->m_R.Delete();
		delete pElement;

// erase returns iterator to next valid entry or to end()
		iterNames = mapNames.erase(iterNames);
	}
	return 1;
}
//==================================================
//==================================================
//==================================================
//
// class xlName
//
//==================================================
//==================================================
//==================================================

int xlName::m_NumInternalNames = 0;
//==================================================
// Create a reference to an existing Excel name
//==================================================
bool xlName::Set(const char *name)
{
	m_Defined = m_RefValid = m_Worksheet = false;

	if(!name || !*name)
		return false;

	m_Worksheet = (strchr(name, '!') != NULL);
	m_RangeName = name;

// If name exists and is internal, add to the list.
// add_name_record() has no effect if already there.
	if(RefreshRef())
	{
		if(!m_Worksheet)
			add_name_record(NULL, *this);

		return true;
	}
	return false;
}
//==================================================
// Set to the corresponding name if it exists
//==================================================
bool xlName::SetToRef(const cpp_xloper &Ref, bool internal)
{
	Clear();

	if(!Ref.IsRef())
		return false;

//-----------------------------------------------------------
// Convert to text of form [Book1.xls]Sheet1!R1C1
//-----------------------------------------------------------
	cpp_xloper RefTextR1C1;
	if(RefTextR1C1.Excel(xlfReftext, 1, &Ref) != xlretSuccess
	|| RefTextR1C1.IsType(xltypeErr))
		return false;

//-----------------------------------------------------------
// Get the name, if it exists, otherwise fail.
//
// First look for an internal name (the default if the 2nd
// argument to xlfGetDef is omitted).
//-----------------------------------------------------------
	if(internal)
	{
		if(m_RangeName.Excel(xlfGetDef, 1, &RefTextR1C1) != xlretSuccess
		|| !m_RangeName.IsType(xltypeStr))
			return m_Defined = m_RefValid = false;

		m_Worksheet = false;
		m_Defined = m_RefValid = true;
// If name exists and is internal, add to the list.
// add_name_record() has no effect if already there.
// Need m_Defined = true before calling this:
		add_name_record(NULL, *this);
	}
	else
	{
// Extract the sheet name and specify this explicitly
		cpp_xloper SheetName;

		if(SheetName.Excel(xlSheetNm, 1, &Ref) != xlretSuccess
		|| !SheetName.IsType(xltypeStr))
			return m_Defined = m_RefValid = false;

// Truncate RefTextR1C1 at the R1C1 part
		char *p = (char *)RefTextR1C1; // need to free this
		RefTextR1C1 = strchr(p, '!') + 1;
		free(p); // free the deep copy

// Truncate SheetName at the sheet name
		p = (char *)SheetName;
		SheetName = strchr(p, ']') + 1;
		free(p); // free the deep copy

		if(m_RangeName.Excel(xlfGetDef, 2, &RefTextR1C1, &SheetName) != xlretSuccess
		|| !m_RangeName.IsType(xltypeStr))
			return m_Defined = m_RefValid = false;

		m_Worksheet = true;
		m_Defined = m_RefValid = true;
	}
	return true;
}
//==================================================
// Set to be the range associated with caller or return
// false if none exists. If dll_only == true, only look for
// an internal worksheet name
//==================================================
bool xlName::SetToCallersName(void)
{
	Clear();

// Get a reference to the calling cell
	cpp_xloper Caller;

	if(Caller.Excel(xlfCaller) != xlretSuccess)
		return false;

	return SetToRef(Caller, true); // true: look for internal name
}
//==================================================
// Clears memory associated with this instance but does not delete range
//==================================================
void xlName::Clear(void)
{
	if(m_Defined)
	{
		m_Defined = m_RefValid = m_Worksheet = false;
		m_RangeName.Free();
		m_RangeRef.Free();
	}
}
//==================================================
void xlName::GetName(cpp_xloper &Name) const
{
	if(!m_Defined)
		return;

	Name = m_RangeName; // Makes a deep copy
}
//==================================================
char *xlName::GetName(void) const
{
	if(!m_Defined)
		return NULL;

	return (char *)m_RangeName; // Returns a deep copy.  Caller must free
}
//==================================================
bool xlName::NameIs(const char *name) const
{
	return (m_RangeName == name);
}
//==================================================
bool xlName::RefreshRef(void)
{
// Update the reference corresponding to m_RangeName
// by aksing Excel to evaluate it using xlfEvaluate.
// The method frees m_RangeRef resources before assigning new value
	if(m_RangeRef.Excel(xlfEvaluate, 1, &m_RangeName) != xlretSuccess
	|| m_RangeRef.IsNameErr()) // Name not defined
		return m_Defined = m_RefValid = false;

	if(!m_RangeRef.IsRef())
	{
		m_Defined = true;
		return m_RefValid = false;
	}
	return m_Defined = m_RefValid = true;
}
//==================================================
char * xlName::GetDef(void) const // get defintion as string (caller must free)
{
	cpp_xloper Defn;

	if(Defn.Excel(xlfGetName, 1, &m_RangeName) != xlretSuccess
	|| !Defn.IsType(xltypeStr))
		return NULL;

	return (char *)Defn;
}
//==================================================
bool xlName::NameCaller(const char *name)
{
//-----------------------------------------------------------
// Check if given internal name already exists for this caller
//-----------------------------------------------------------
	if(SetToCallersName() && !m_Worksheet)
	{
// If no name specified, then the existing name is what's required
		if(!name || !*name)
			return true;

// Check if name is the same as the specified one
		if(m_RangeName == name)
			return true;

// If not, delete the old name, create a new one.
		Delete();
	}

//-----------------------------------------------------------
// If no name provided, create a unique name
//-----------------------------------------------------------
	if(!name || !*name)
	{
		char *p = make_unique_name();
		m_RangeName = p;
		free(p);
	}
	else
	{
		m_RangeName = name;
	}
	m_Worksheet = false;  // This will be an internal name

//-----------------------------------------------------------
// Get a reference to the calling cell
//-----------------------------------------------------------
	cpp_xloper Caller;

	if(Caller.Excel(xlfCaller) != xlretSuccess)
		return m_Defined = m_RefValid = false;

//-----------------------------------------------------------
// Associate the new internal name with the calling cell(s)
//-----------------------------------------------------------
	cpp_xloper RetVal;

	if(RetVal.Excel(xlfSetName, 2, &m_RangeName, &Caller) != xlretSuccess)
		return m_Defined = m_RefValid = false;

//-----------------------------------------------------------
// Add the new internal name to the list
//-----------------------------------------------------------
	m_Defined = m_RefValid = true;
	add_name_record(NULL, *this);
	return true;
}
//==================================================
// If defined, deletes the given name and frees resources
//==================================================
void xlName::Delete(void)
{
	RefreshRef();
	cpp_xloper RetVal;

	if(m_Defined) // Don't care if ref invalid.  Still delete the name
	{
		if(m_Worksheet) // Can only do this in a command
			RetVal.Excel(xlcDeleteName, 1, &m_RangeName);
		else // Internal names can be deleted any time
			RetVal.Excel(xlfSetName, 1, &m_RangeName);
	}
	Clear();
}
//==================================================
// Change (or define for the first time) the defition
// of the range name already associated with this
// instance of xlName
//==================================================
bool xlName::Define(const cpp_xloper &Defintion, bool in_dll)
{
// Delete any existing definition for this range name
	Delete();

	m_Worksheet = !in_dll;

// Create a new definition
	cpp_xloper RetVal;
	bool rtn_val;

	if(in_dll) // a hidden internal dll name
	{
		rtn_val = RetVal.Excel(xlfSetName, 2, &m_RangeName, &Defintion) == xlretSuccess;
// Add to the list of DLL-internal names, enabling clear up
		add_name_record(NULL, *this);
	}
	else // a worksheet name.  Can only do this in a command
	{
		rtn_val = RetVal.Excel(xlcDefineName, 2, &m_RangeName, &Defintion) == xlretSuccess;
	}

	if(!rtn_val || RetVal.IsType(xltypeErr))
		return m_Defined = m_RefValid = false;

	return m_Defined = m_RefValid = true;
}
//==================================================
// Create (or define for the first time) a named range
// and associate it with this instance of xlName
bool xlName::Define(const char *name, const cpp_xloper &Defintion, bool in_dll)
{
	if(!name || !*name)
		return false;

	m_RangeName = name;

	return Define(Defintion, in_dll);
}
//==================================================
xlName & xlName::operator=(const xlName &source)
{
	if(&source == this)
		return *this;

	m_Defined = source.m_Defined;
	m_RefValid = source.m_RefValid;
	m_Worksheet = source.m_Worksheet;

// cpp_xloper assignment operator makes deep copies
	m_RangeRef = source.m_RangeRef;
	m_RangeName = source.m_RangeName;
	return *this;
}
//==================================================
// Overloaded operators
//==================================================
// Assignment operators place values in cell(s) that
// range refers to.  Cast operators retrieve values
// or assign nil if range is not valid or conversion
// was not possible.  Casting to char * will (and
// xloper * can) return dynamically allocated memory
// that the caller must free.
//
// Retrieving values is easier than setting them, as
// it is not necessary to refresh the reference first
//==================================================
int xlName::operator=(int i)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Temp((double)i);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Temp);
	}
	return i;
}
//==================================================
bool xlName::operator=(bool b)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Value(b);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Value);
	}
	return b;
}
//==================================================
double xlName::operator=(double d)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Temp(d);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Temp);
	}
	return d;
}
//==================================================
WORD xlName::operator=(WORD xl_error)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Value(xl_error);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Value);
	}
	return xl_error;
}
//==================================================
const char *xlName::operator=(const char *s)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Value(s);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Value);
	}
	return s;
}
//==================================================
const xloper *xlName::operator=(const xloper *p_op)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Op(p_op);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Op);
	}
	return p_op;
}
//==================================================
const xloper12 *xlName::operator=(const xloper12 *p_op)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Op(p_op);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Op);
	}
	return p_op;
}
//==================================================
const VARIANT *xlName::operator=(const VARIANT *p_var)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Value(p_var);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Value);
	}
	return p_var;
}
//==================================================
const xl4_array *xlName::operator=(const xl4_array *array)
{
	if(RefreshRef())
	{
		cpp_xloper RetVal, Value(array->rows, array->columns, array->array);
		RetVal.Excel(xlSet, 2, &m_RangeRef, &Value);
	}
	return array;
}
//==================================================
double xlName::operator+=(double d)
{
	if(!RefreshRef())
		return 0.0;

	cpp_xloper RetVal, Value;
	cpp_xloper Type(xltypeNum);

	if(Value.Excel(xlCoerce, 2, &m_RangeRef, &Type) != xlretSuccess || !Value.IsNum())
		return 0.0;

	Value += d;
	RetVal.Excel(xlSet, 2, &m_RangeRef, &Value);
	return (double)Value;
}
//==================================================
// Retrieving values is easier than setting them, as
// it is not necessary to refresh the reference first
//==================================================
xlName::operator int(void) const
{
	cpp_xloper Value;
	if(Value.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Type(xltypeNum);
		if(Value.Excel(xlCoerce, 2, &Value, &Type) == xlretSuccess)
			return (int)Value;
	}
	return 0;
}
//==================================================
xlName::operator bool(void) const
{
	cpp_xloper Value;
	if(Value.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Type(xltypeBool);
		if(Value.Excel(xlCoerce, 2, &Value, &Type) == xlretSuccess)
			return (bool)Value;
	}
	return false;
}
//==================================================
xlName::operator double(void) const
{
	cpp_xloper Value;
	if(Value.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Type(xltypeNum);
		if(Value.Excel(xlCoerce, 2, &Value, &Type) == xlretSuccess)
			return (double)Value;
	}
	return 0.0;
}
//==================================================
xlName::operator char *(void) const // returns range contents as string
{
	cpp_xloper Value;
	if(Value.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Type(xltypeStr);
		if(!Value.ConvertToString())
			return (char *)Value;
	}
	return NULL;
}
//==================================================
xlName::operator wchar_t *(void) const // DLL-allocated copy, caller must free
{
	cpp_xloper Value;
	if(Value.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Type(xltypeStr);
		if(!Value.ConvertToString())
			return (wchar_t *)Value;
	}
	return NULL;
}
//==================================================
// Get the size of the range
bool xlName::GetRangeSize(DWORD &size) const
{
	cpp_xloper Range;

	if(Range.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Values;
		cpp_xloper Type(xltypeMulti);

		if(Values.Excel(xlCoerce, 2, &Range, &Type) == xlretSuccess && Values.IsMulti())
			return Values.GetArraySize(size);
	}
	return false;
}
//==================================================
bool xlName::GetRangeSize(RW &rows, COL &columns) const
{
	cpp_xloper Range;

	if(Range.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Values;
		cpp_xloper Type(xltypeMulti);

		if(Values.Excel(xlCoerce, 2, &Range, &Type) == xlretSuccess && Values.IsMulti())
			return Values.GetArraySize(rows, columns);
	}
	return false;
}
//==================================================
// Retrieve the named range contents as xltypeMulti
// This is a two-stage process:
// 1. Convert the text name to a ref using xlfEvaluate
// 2. Convert the ref to xltypeMulti using xlCoerce
bool xlName::GetValues(cpp_xloper &Values) const
{
	Values.Free();

	cpp_xloper Range;

	if(Range.Excel(xlfEvaluate, 1, &m_RangeName) == xlretSuccess)
	{
		cpp_xloper Type(xltypeMulti);
		return (Values.Excel(xlCoerce, 2, &Range, &Type) == xlretSuccess && Values.IsMulti());
	}
	return false;
}
//==================================================
bool xlName::GetRef(cpp_xloper &Ref) const // range as xltypeRef
{
	Ref.Free();
	Ref = m_RangeRef; // Deep copy
	return true;
}
//==================================================
// Retrieve at most array_size doubles from the range
// This is a three-stage process:
// 1. Convert the text name to a ref using xlfEvaluate
// 2. Convert the ref to xltypeMulti using xlCoerce
// 3. Convert the xltypeMulti to doubles
bool xlName::GetValues(double *array, DWORD array_size) const
{
	if(!array || !array_size)
		return false;

	cpp_xloper Range;

	if(Range.Excel(xlfEvaluate, 1, &m_RangeName) != xlretSuccess)
		return false;

	cpp_xloper Values;
	cpp_xloper Type(xltypeMulti);

	if(Values.Excel(xlCoerce, 2, &Range, &Type) != xlretSuccess
	|| !Values.IsMulti())
		return false;

	DWORD num_values;
	Values.GetArraySize(num_values);

	double *p = Values.ConvertMultiToDouble();
	if(!p)
		return false;

	if(num_values > array_size)
		num_values = array_size;

	memcpy(array, p, sizeof(double) * num_values);
	free(p);
	return true;
}
//==================================================
bool xlName::SetValues(const cpp_xloper &Values)
{
// Assign Values to named range
	cpp_xloper RetVal;
	return RefreshRef() && (RetVal.Excel(xlSet, 2, &m_RangeRef, &Values) == xlretSuccess);
}
//==================================================
bool xlName::SetValues(const double *array, RW rows, COL columns)
{
	cpp_xloper RetVal, Values(rows, columns, array);
	return RefreshRef() && (RetVal.Excel(xlSet, 2, &m_RangeRef, &Values) == xlretSuccess);
}
//==================================================
void xlName::SetNote(const char *text)
{
	if(!RefreshRef())
		return;

	cpp_xloper RetVal;

	if(text)
	{
		cpp_xloper Note(text);
		RetVal.Excel(xlcNote, 2, &Note, &m_RangeRef);
	}
	else
	{
		cpp_xloper Missing(p_xlMissing);
		RetVal.Excel(xlcNote, 2, &Missing, &m_RangeRef);
	}
}
//==================================================
char * xlName::GetNote(void)
{
	if(!RefreshRef())
		return NULL;

	cpp_xloper Note;

	if((Note.Excel(xlcNote, 1, &m_RangeRef) == xlretSuccess) && Note.IsStr())
		return (char *)Note;

	return NULL;
}
//==================================================
// End of xlName code
//==================================================


//========================================================
//========================================================
//========================================================
//========================================================
unsigned long now_serial_seconds(void)
{
	time_t time_t_T;
	time(&time_t_T);
	tm tm_T = *localtime(&time_t_T);
	return (unsigned long)tm_T.tm_sec
		+ 60 * (tm_T.tm_min
		+ 60 * (tm_T.tm_hour
		+ 24 * (tm_T.tm_yday
		+ 366 * tm_T.tm_year % 100)));
}

// The following class generates unique names in a way
// that is safe to be called from MTR versions of Excel
// such as Excel 2007

class UniqueNameFactory
{
public:
	UniqueNameFactory(void)
	{
		if(gExcelVersion12plus)
			InitializeCriticalSection(&m_CS);
		m_Count = 0;
		m_Tlast = 0;
	}
	~UniqueNameFactory(void)
	{
		if(gExcelVersion12plus)
			DeleteCriticalSection(&m_CS);
	}

	char *GetNewName(void)
	{
	// Need an unsigned long to contain max possible value
		unsigned long T = now_serial_seconds();

		if(gExcelVersion12plus)
			EnterCriticalSection(&m_CS);

		if(T != m_Tlast)
		{
			m_Tlast = T;
			m_Count = 0;
		}
		else
		{
			++m_Count;
		}
		char buffer[32]; // More than enough space

	// Increment name_count so that names created in the current
	// second are still unique.  The name_count forms the first
	// part of the name.
		int ch_count = sprintf(buffer, "x%ld.", m_Count);

		if(gExcelVersion12plus)
			LeaveCriticalSection(&m_CS);

		int r;
	// Represent the time number in base 62 using 0-9, A-Z, a-z.
	// Puts the characters most likely to differ at the front
	// of the name to optimise name searches and comparisons
		for(;T; T /= 62)
		{
			if((r = T % 62) < 10)
				r += '0';
			else if(r < 36)
				r += 'A' - 10;
			else
				r += 'a' - 36;

			buffer[ch_count++] = r;
		}
		buffer[ch_count] = 0;

	// Make a copy of the string and return it
		char *new_name = (char *)malloc(ch_count + 1);
		strcpy(new_name, buffer);
		return new_name; // caller must free the memory
	}

private:
	long m_Count;
	unsigned long m_Tlast; 
	CRITICAL_SECTION m_CS;
};

// This function is not thread-safe
char *make_unique_name(void)
{
	static long name_count = 0;
	static unsigned long T_last = 0;

// Need an unsigned long to contain max possible value
	unsigned long T = now_serial_seconds();

	if(T != T_last)
	{
		T_last = T;
		name_count = 0;
	}

	char buffer[32]; // More than enough space

// Increment name_count so that names created in the current
// second are still unique.  The name_count forms the first
// part of the name.
	int ch_count = sprintf(buffer, "x%ld.", ++name_count);

	int r;
// Represent the time number in base 62 using 0-9, A-Z, a-z.
// Puts the characters most likely to differ at the front
// of the name to optimise name searches and comparisons
	for(;T; T /= 62)
	{
		if((r = T % 62) < 10)
			r += '0';
		else if(r < 36)
			r += 'A' - 10;
		else
			r += 'a' - 36;

		buffer[ch_count++] = r;
	}
	buffer[ch_count] = 0;

// Make a copy of the string and return it
	char *new_name = (char *)malloc(ch_count + 1);
	strcpy(new_name, buffer);
	return new_name; // caller must free the memory
}
