//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================
// This source file contains examples functions that manipulate text strings.
//============================================================================
//============================================================================

#include <windows.h>

#include "XllAddIn.h"

//===================================================================
char *deep_copy_xlstring(const char *xlstring)
{
	if(!xlstring)
		return NULL;

	size_t len = (BYTE)xlstring[0] + 2;
	char *p = (char *)malloc(len);
	memcpy(p, xlstring, len); // copies the char count, string and null termination
	return p;
}
//-------------------------------------------------------------------
// Deep copy and cast down from Unicode to ASCII byte string
char *deep_copy_xlstring(const wchar_t *xl12string)
{
	if(!xl12string)
		return NULL;

	size_t len = xl12string[0];
	char *p = (char *)malloc(len + 2);
	if(!p || (len = wcstombs(p + 1, xl12string + 1, xl12string[0])) < 0)
	{
		free(p);
		return NULL;
	}
	p[0] = len; // string p[1] is NOT necessarily null terminated
	p[len + 1] = 0; // now it is
	return p;
}
//-------------------------------------------------------------------
// Deep copy and cast up from ASCII to Unicode string
wchar_t *deep_copy_xl12string(const char *xlstring)
{
	if(!xlstring)
		return NULL;

	size_t len, mem_size = ((BYTE)xlstring[0] + 2) * sizeof(wchar_t);
	wchar_t *p = (wchar_t *)malloc(mem_size);
	if((len = mbstowcs(p + 1, xlstring + 1, (BYTE)xlstring[0])) == -1)
	{
		free(p);
		return NULL;
	}
	p[0] = len; // string p[1] is NOT necessarily null terminated
	p[len + 1] = 0; // now it is
	return p;
}
//-------------------------------------------------------------------
wchar_t *deep_copy_xl12string(const wchar_t *xl12string)
{
	if(!xl12string)
		return NULL;

	size_t mem_size = (xl12string[0] + 2) * sizeof(wchar_t);
	wchar_t *p = (wchar_t *)malloc(mem_size);
	memcpy(p, xl12string, mem_size); // copies the char count, string and null termination
	return p;
}
//-------------------------------------------------------------------
// Create counted ASCII byte string from null-terminated ASCII input
char *new_xlstring(const char *text)
{
	size_t len;

	if(!text || !(len = strlen(text)))
		return NULL;

	if(len > MAX_XL4_STR_LEN)
		len = MAX_XL4_STR_LEN; // truncate

	char *p = (char *)malloc(len + 2);
	if(!p) return NULL;
	memcpy(p + 1, text, len + 1);
	p[0] = (BYTE)len;
	return p;
}
//-------------------------------------------------------------------
// Create counted ASCII byte string from null-terminated Unicode input
char *new_xlstring(const wchar_t *text)
{
	size_t len;

	if(!text || !(len = wcslen(text)))
		return NULL;

	if(len > MAX_XL4_STR_LEN)
		len = MAX_XL4_STR_LEN; // truncate

	char *p = (char *)malloc(len + 2);
	if(!p || wcstombs(p + 1, text, len) < 0)
	{
		free(p);
		return NULL;
	}
	p[0] = (BYTE)len;
	p[len + 1] = 0;
	return p;
}
//-------------------------------------------------------------------
// Create counted Unicode wchar string from null-terminated ASCII input
wchar_t *new_xl12string(const char *text)
{
	size_t len;

	if(!text || !(len = strlen(text)))
		return NULL;

	if(len > MAX_XL12_STR_LEN)
		len = MAX_XL12_STR_LEN; // truncate

	wchar_t *p = (wchar_t *)malloc((len + 2) * sizeof(wchar_t));
	if(!p) return NULL;
	mbstowcs(p + 1, text, len);
	p[0] = len; // string p[1] is NOT null terminated
	p[len + 1] = 0; // now it is
	return p;
}
//-------------------------------------------------------------------
// Create counted Unicode wchar string from null-terminated Unicode input
wchar_t *new_xl12string(const wchar_t *text)
{
	size_t len;

	if(!text || !(len = wcslen(text)))
		return NULL;

	if(len > MAX_XL12_STR_LEN)
		len = MAX_XL12_STR_LEN; // truncate

	wchar_t *p = (wchar_t *)malloc((len + 2) * sizeof(wchar_t));
	if(!p) return NULL;
	memcpy(p + 1, text, (len + 1) * sizeof(wchar_t));
	p[0] = len;
	return p;
}
//-------------------------------------------------------------------
// Returns a new xlstring which the caller must free.  Does not
// free the original string.
//-------------------------------------------------------------------
char *append_xlstring(const char *xlstring, const char *new_text)
{
	size_t new_len = strlen(new_text), old_len = (BYTE)xlstring[0], len = old_len + new_len;

	if(!new_text || !new_len || old_len == MAX_XL4_STR_LEN)
		return NULL;

	if(len > MAX_XL4_STR_LEN)
	{
		len = MAX_XL4_STR_LEN;
		new_len = old_len - len; // prevents buffer overrun
	}

	char *new_buffer = (char *)malloc(len + 2);

	if(!new_buffer)
		return NULL; // No effect

	strncpy(new_buffer + 1, xlstring + 1, old_len);
	strncpy(new_buffer + old_len + 1, new_text, new_len);
	new_buffer[0] = (BYTE)len;
	new_buffer[len + 1] = 0;
	return new_buffer;
}
//-------------------------------------------------------------------
char *append_xlstring(const char *xlstring, const wchar_t *new_text)
{
	size_t new_len = wcslen(new_text), old_len = (BYTE)xlstring[0], len = old_len + new_len;

	if(!new_text || !new_len || old_len == MAX_XL4_STR_LEN)
		return NULL;

	if(len > MAX_XL4_STR_LEN)
	{
		len = MAX_XL4_STR_LEN;
		new_len = old_len - len; // prevents buffer overrun
	}

	char *new_buffer = (char *)malloc(len + 2);
	char *temp_buffer = (char *)malloc(new_len);

	if(!new_buffer || !temp_buffer)
	{
		free(new_buffer);
		free(temp_buffer);
		return NULL; // No effect
	}
// Convert from wchar_t string to multi-byte string
	wcstombs(temp_buffer, new_text, new_len);

	strncpy(new_buffer + 1, xlstring + 1, old_len);
	strncpy(new_buffer + old_len + 1, temp_buffer, new_len);
	new_buffer[0] = len;
	new_buffer[len + 1] = 0;
	free(temp_buffer);
	return new_buffer;
}
//-------------------------------------------------------------------
wchar_t *append_xl12string(const wchar_t *xl12string, const wchar_t *new_text)
{
	size_t new_len = wcslen(new_text), old_len = xl12string[0], len = old_len + new_len;

	if(!new_text || !new_len || old_len == MAX_XL12_STR_LEN)
		return NULL;

	if(len > MAX_XL12_STR_LEN)
	{
		len = MAX_XL12_STR_LEN;
		new_len = old_len - len; // prevents buffer overrun
	}

	wchar_t *new_buffer = (wchar_t *)malloc((len + 2) * sizeof(wchar_t));

	if(!new_buffer)
		return NULL; // No effect

	wcsncpy(new_buffer + 1, xl12string + 1, old_len);
	wcsncpy(new_buffer + old_len + 1, new_text, new_len);
	new_buffer[0] = len;
	new_buffer[len + 1] = 0;
	return new_buffer;
}
//-------------------------------------------------------------------
wchar_t *append_xl12string(const wchar_t *xl12string, const char *new_text)
{
	size_t new_len = strlen(new_text), old_len = xl12string[0], len = old_len + new_len;

	if(!new_text || !new_len || old_len == MAX_XL12_STR_LEN)
		return NULL;

	if(len > MAX_XL12_STR_LEN)
	{
		len = MAX_XL12_STR_LEN;
		new_len = old_len - len; // prevents buffer overrun
	}

	wchar_t *new_buffer = (wchar_t *)malloc((len + 2) * sizeof(wchar_t));
	wchar_t *temp_buffer = (wchar_t *)malloc(new_len * sizeof(wchar_t));

	if(!new_buffer || !temp_buffer)
	{
		free(new_buffer);
		free(temp_buffer);
		return NULL; // No effect
	}
// Convert to wchar_t string from multi-byte string
	mbstowcs(temp_buffer, new_text, new_len);

	wcsncpy(new_buffer + 1, xl12string + 1, old_len);
	wcsncpy(new_buffer + old_len + 1, temp_buffer, new_len);
	new_buffer[0] = len;
	new_buffer[len + 1] = 0;
	free(temp_buffer);
	return new_buffer;
}
//-------------------------------------------------------------------
wchar_t *concat_xlstrings(const wchar_t *xlstring1, const wchar_t *xlstring2)
{
	if(!xlstring1 || !xlstring2)
		return NULL;

	size_t len1 = xlstring1[0], len2 = xlstring2[0];

	if(len1 > MAX_XL12_STR_LEN)
	{
		len1 = MAX_XL12_STR_LEN;
		len2 = 0;
	}

	int new_len = len1 + len2;

	if(new_len > MAX_XL12_STR_LEN)
	{
		len2 = MAX_XL12_STR_LEN - len1;
		new_len = MAX_XL12_STR_LEN;
	}
	wchar_t *new_buffer = (wchar_t *)malloc((new_len + 2) * sizeof(wchar_t));

	if(!new_buffer)
		return NULL; // No effect

	wcsncpy(new_buffer + 1, xlstring1 + 1, len1);
	if(len2) wcsncpy(new_buffer + 1 + len1, xlstring2 + 1, len2);
	new_buffer[0] = new_len;
	new_buffer[new_len + 1] = 0;
	return new_buffer;
}
//-------------------------------------------------------------------
char *concat_xlstrings(const char *xlstring1, const char *xlstring2)
{
	if(!xlstring1 || !xlstring2)
		return NULL;

	size_t len1 = (BYTE)xlstring1[0], len2 = (BYTE)xlstring2[0];

	if(len1 > MAX_XL4_STR_LEN)
	{
		len1 = MAX_XL4_STR_LEN;
		len2 = 0;
	}

	int new_len = len1 + len2;

	if(new_len > MAX_XL4_STR_LEN)
	{
		len2 = MAX_XL4_STR_LEN - len1;
		new_len = MAX_XL4_STR_LEN;
	}
	char *new_buffer = (char *)malloc(new_len + 2);

	if(!new_buffer)
		return NULL; // No effect

	strncpy(new_buffer + 1, xlstring1 + 1, len1);
	if(len2) strncpy(new_buffer + 1 + len1, xlstring2 + 1, len2);
	new_buffer[0] = new_len;
	new_buffer[new_len + 1] = 0;
	return new_buffer;
}
//===================================================================
size_t count_char(char *text, char ch)
{
	if(!text || !ch)
		return 0;
	for(size_t count = 0; *text; )
		if(*text++ == ch)
			count++;
	return count;
}
//===================================================================
size_t count_char(wchar_t *text, wchar_t ch)
{
	if(!text || !ch)
		return 0;
	for(size_t count = 0; *text; )
		if(*text++ == ch)
			count++;
	return count;
}
//========================================================
size_t count_chars(char *text, char *chars)
{
	if(!text || !chars)
		return 0;

	short count = 0;
	char *p;

	do
	{
		for(p = chars; *p;)
		{
			if(*text == *p++)
			{
				count++;
				break;
			}
		}
	}
	while(*(++text));

	return count;
}
//========================================================
// Returns pointer to the first character in
// text that does not belong to the set of characters in search_text.
// Returns NULL if text consists entirely of characters from string2 or if
// there was an error.
char *strspnp(char *text, char *search_text)
{
	if(!text || !search_text)
		return NULL;

	for(char *t = text; *t; t++)
		if(!strchr(search_text, *t)) // then *t was not in search_text, so return its position
			return t;

	return NULL; // all of text chars are in search_text (but not necessarily vice versa)
}
