//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================

//============================================================================
//============================================================================
// This source file (and header file) contain code examples relating to the
// creation of a background thread in the DLL and the creation and management
// of tasks associated with worksheet cells.  These tasks are run in
// background by the thread.  When tasks are complete the worksheet cells are
// updated under the control of a repeating command function associated with
// this task thread.
//============================================================================
//============================================================================
#include <windows.h>
#include <stdio.h>
#include <time.h>

#include "XllAddIn.h"
#include "XllNames.h"
#include "Background.h"
#include "TLS.h"

char *make_unique_name(void);

#define DEFAULT_THREAD_SLEEP_MS		250
#define THREAD_INACTIVE_SLEEP_MS	1000

int __stdcall LongTaskPollingCmd(void);

TaskList ExampleTaskList(LongTaskPollingCmd, 5); // poll every 5 secs

DWORD __stdcall background_thread_main(void *vp);

//============================================================================
//============================================================================

// task member functions

//============================================================================
//============================================================================
//============================================================================
task::task()
{
	clear();
}
//============================================================================
task::task(int n_args, const cpp_xloper *InArray) : num_args(n_args)
{
	clear();
	if(num_args > 0)
	{
		ArgArray = new cpp_xloper[num_args];
		if(InArray)
			for(int i = num_args; i--; ArgArray[i] = InArray[i]);
	}
	FnRetVal.SetToError(xlerrNum); // default 'pending' value
}
//============================================================================
task::~task()
{
	if(num_args)
	{
		for(int i = num_args; i--; ArgArray[i].Free());
		delete[] ArgArray;
	}
	free(caller_name);
	clear();
}
//============================================================================
void task::clear(void)
{
	prev = next = NULL;
	ArgArray = NULL;
	num_args = 0;
	caller_name = NULL;
}
//============================================================================
void task::modify_args(const cpp_xloper *InArray, int n_args)
{
// Free the old arguments and set the new
	if(num_args)
	{
		for(int i = num_args; i--; ArgArray[i].Free());
		delete[] ArgArray;
	}
	num_args = n_args;

	if(num_args > 0)
	{
		ArgArray = new cpp_xloper[num_args];
		if(InArray)
			for(int i = num_args; i--; ArgArray[i] = InArray[i]);
	}
	FnRetVal.SetToError(xlerrNum); // default 'pending' value
	status = TASK_PENDING;
}
//============================================================================
bool task::args_changed(const cpp_xloper *InArray, int n_args) const
{
	if(num_args != n_args)
		return true;

	for(int i = num_args; i--;)
		if(!(ArgArray[i] == InArray[i]))
			return true;

	return false; // nothing is different
}
//============================================================================
//============================================================================

// TaskList member functions

//============================================================================
//============================================================================
//============================================================================
//
//	THESE FUNCTIONS ARE ONLY CALLED BY THE FOREGROUND THREAD
//
//============================================================================
//============================================================================
//============================================================================
TaskList::TaskList(int (__stdcall * Cmd)(void), short FreqSecs)
{
	m_pTail = m_pHead = m_pCurrent = NULL;

	if(!Cmd || FreqSecs < 1)
		return;

	InitializeCriticalSection(&m_ListControl);
	InitializeCriticalSection(&m_ThreadControl);

	m_ThreadExitFlagSet = m_SuspendAllFlagSet = false;
	m_ThreadIsCreated = m_ThreadIsRunning = false;

	m_ThreadHandle = 0;
	m_ThreadID = 0;
	m_ThreadSleepMs = DEFAULT_THREAD_SLEEP_MS;

	m_BreakPollingCmdFlag = false;
	PollingCmd = Cmd;
	m_PollingCmdFreqSecs = FreqSecs;
}
//============================================================================
TaskList::~TaskList(void)
{
// Free all the memory associated with the tasks and terminate the thread
	if(m_ThreadIsCreated)
		DeleteTaskThread();

	DeleteCriticalSection(&m_ThreadControl);
	DeleteCriticalSection(&m_ListControl);
}
//============================================================
bool TaskList::CreateTaskThread(void)
{
	if(m_ThreadIsCreated)
		return true;

// Start the OS thread in an active state, but stop it from
// doing anything for now. Pass this instance of the class
// as a parameter so that background_thread_main() can call
// this instance's own main function, TaskThreadMain
	m_ThreadHandle = CreateThread(NULL, 0, background_thread_main, this, 0, &m_ThreadID);

	if(!m_ThreadHandle)
		return false;

//	m_ThreadIsRunning = false; // should already have this value
	m_ThreadIsCreated = true;
	return true;
}
//============================================================
void TaskList::DeleteTaskThread(void) // exits the thread and reinitialises
{
	if(!m_ThreadIsCreated)
		return;

// Stop the polling command from executing its next cycle
	m_BreakPollingCmdFlag = true;

// First suspend the thread. This takes care of stopping
// the current task.  The thread is still active but will
// not process any tasks
	SuspendTaskExecution();

// Tell background thread to return (i.e. exit)
	EnterCriticalSection(&m_ThreadControl);
	m_ThreadExitFlagSet = true;
	LeaveCriticalSection(&m_ThreadControl);

// Wait for the thread to terminate.
	DWORD Code;
	for(;GetExitCodeThread(m_ThreadHandle, &Code) && Code == STILL_ACTIVE; Sleep(10));

// Delete the thread object by releasing the handle
	CloseHandle(m_ThreadHandle);

// Delete all the tasks and free related resources
	DeleteAllTasks();

	m_SuspendAllFlagSet = m_ThreadExitFlagSet = m_ThreadIsCreated = false;
	m_ThreadHandle = 0;
	m_ThreadID = 0;
}
//============================================================
bool TaskList::SuspendTaskExecution(void)
{
	if(!m_ThreadIsCreated || !m_ThreadIsRunning)
		return true;

// Stop the polling command from executing its next cycle
	m_BreakPollingCmdFlag = true;

	EnterCriticalSection(&m_ThreadControl);
	m_SuspendAllFlagSet = true;
	LeaveCriticalSection(&m_ThreadControl);

	EnterCriticalSection(&m_ListControl);
	if(m_pCurrent) // flag current task execution to break
		m_pCurrent->break_task = true;
	LeaveCriticalSection(&m_ListControl);

	for(; m_ThreadIsRunning; Sleep(10));
	return true;
}
//============================================================
bool TaskList::ResumeTaskExecution(void) // restarts suspended thread
{
	if(!m_ThreadIsCreated)
		return false;
		
	if(m_ThreadIsRunning)
		return true;

	m_SuspendAllFlagSet = false;
	m_ThreadIsRunning = true;

	EnterCriticalSection(&m_ListControl);
	if(m_pCurrent) // flag current task execution to resume
		m_pCurrent->break_task = false;
	LeaveCriticalSection(&m_ListControl);

// Restart the polling command cycle
	m_BreakPollingCmdFlag = false;
	PollingCmd();
	return true;
}
//============================================================================
bool TaskList::SetThreadSleepMs(long sleep_ms)
{
	if(sleep_ms < 0)
		return false;
	m_ThreadSleepMs = sleep_ms;
	return true;
}
//============================================================================
bool TaskList::SetPollingSecs(short rpt_secs)
{
	if(rpt_secs < 1)
		return false;
	m_PollingCmdFreqSecs = rpt_secs;
	return true;
}
//============================================================================
//============================================================================
// Task list management
//============================================================================
//============================================================================
void TaskList::GetTaskStats(long &pending, long &done)
{
	pending = done = 0;
	EnterCriticalSection(&m_ListControl);
	for(task *p = m_pHead; p; p = p->next)
	{
		if(p->status == TASK_PENDING)
			pending++;
		else if(p->status & (TASK_READY | TASK_UNCLAIMED | TASK_COMPLETE))
			done++;
	}
	LeaveCriticalSection(&m_ListControl);
}
//============================================================================
task *TaskList::FindTask(char *name) const
{
	for(task *p = m_pHead; p; p = p->next)
		if(strcmp(name, p->caller_name) == 0)
			return p;
	return NULL;
}
//============================================================================
bool TaskList::RemoveTaskFromList(task *p)
{
// Need to enter m_ListControl as we may be changing things
// that the background thread is reading or might read.
	EnterCriticalSection(&m_ListControl);

	if(!m_pHead || !p || (m_ThreadIsRunning && p == m_pCurrent))
	{
		LeaveCriticalSection(&m_ListControl);
		return false;
	}

	if(p == m_pHead)
	{
		if(p->next)
		{
			m_pHead = p->next;
			m_pHead->prev = NULL;
		}
		else // was the only thing in the list
		{
			m_pHead = m_pTail = m_pCurrent = NULL;
		}
	}
	else if(p == m_pTail)
	{
		m_pTail = p->prev; // can't be NULL as m_pHead not NULL
		m_pTail->next = NULL;
	}
	else
	{
		p->next->prev = p->prev; // } could be NULL if was
		p->prev->next = p->next; // } next to m_pHead or m_pTail
	}

	LeaveCriticalSection(&m_ListControl);
	return true;
}
//============================================================================
void TaskList::AddTaskToListTail(task *p)
{
// Need to enter m_ListControl as we may be changing things
// that the background thread is reading.
	EnterCriticalSection(&m_ListControl);
	if(m_pTail) // already something in queue
	{
		m_pTail->next = p;
		p->prev = m_pTail;
		m_pTail = p;
		p->next = NULL;
	}
	else  // this will be the first task to be processed
	{
		m_pTail = m_pHead = p;
		p->prev = p->next = NULL;
	}
	LeaveCriticalSection(&m_ListControl);
}
//============================================================================
bool TaskList::MoveTaskToTail(task *p)
{
// Need to enter m_ListControl as we may be changing things
// that the background thread is reading or might read.
	EnterCriticalSection(&m_ListControl);

	if(!m_pHead || !p || (m_ThreadIsRunning && p == m_pCurrent))
	{
		LeaveCriticalSection(&m_ListControl);
		return false;
	}

	if(p == m_pTail) // nothing to do
	{
		LeaveCriticalSection(&m_ListControl);
		return true;
	}

// Snip the task out of the list
	if(p == m_pHead)
	{
		if(p->next)
		{
			m_pHead = p->next;
			m_pHead->prev = NULL;
		}
		else // was the only thing in the list
		{
			m_pHead = m_pTail = m_pCurrent = NULL;
		}
	}
	else
	{
		p->next->prev = p->prev; // } could be NULL if was
		p->prev->next = p->next; // } next to m_pHead or m_pTail
	}
// Add the task back at the tail
	if(m_pTail) // already something in queue
	{
		m_pTail->next = p;
		p->prev = m_pTail;
		m_pTail = p;
		p->next = NULL;
	}
	else  // this will be the first task to be processed
	{
		m_pTail = m_pHead = p;
		p->prev = p->next = NULL;
	}
	LeaveCriticalSection(&m_ListControl);
	return true;
}
//============================================================================
bool TaskList::MakeNewTask(bool (* fn_ptr)(task *), const cpp_xloper *InArray, short num_args)
{
	xlName Caller;
// Create an internal name and apply it to the calling cell
	if(!Caller.NameCaller(NULL))
		return false;

	char *name = Caller.GetName();

// Create a new task and fill in the data
	task *p = new task(num_args, InArray);

	if(!p)
	{
		free(name);
		return false;
	}
	p->caller_name = name; // freed when task is deleted
	p->status = TASK_PENDING;
	p->fn_ptr = fn_ptr;
	AddTaskToListTail(p);
	return true;
}
//============================================================================
void TaskList::DeleteTask(task *pTask) // frees all task resources
{
	if(!RemoveTaskFromList(pTask))
		return;

	cpp_xloper Name(pTask->caller_name);
	Name.Excel(xlfSetName); // delete the name definition
	pTask->~task();
	delete pTask;
}
//============================================================================
void TaskList::DeleteAllTasks(void)
{
	if(!m_pHead)
		return;

	task *p = m_pHead, *next_p;

	do
	{
		next_p = p->next;
		DeleteTask(p); // frees all task resources
	}
	while((p = next_p));
}
//============================================================================
void TaskList::DeleteUnclaimedTasks(void)
{
	if(!m_pHead)
		return;

	task *p = m_pHead, *next_p;

	do
	{
		next_p = p->next;

		if(p->status == TASK_UNCLAIMED)
			DeleteTask(p); // frees all task resources
	}
	while((p = next_p));
}
//============================================================================
// Break execution of this task if it is the current one
void TaskList::BreakCurrentTask(task *p)
{
	if(!m_pCurrent || p != m_pCurrent)
	{
//	No task to break, or m_pCurrent already changed by background thread
		return;
	}
	p->break_task = true; // this may already have been set, but no harm done

// Wait for status to change.  Note that once broken, the background thread
// will move onto next task unless m_SuspendAllFlagSet is also true.  Note
// that status tested is p->status not m_pCurrent->status.  Note that break
// flag is left set so that this task does not become current again till
// we're ready.
	for(;p->status == TASK_CURRENT; Sleep(10));

// Status of task should now be TASK_PENDING again and m_pCurrent will either
// be NULL or point to the next task to be executed after this one.
}
//============================================================================
// TaskList::UpdateTask performs 3 functions:
//
// 1. Supplies new task to list or updates arguments to existing task
//
// 2. Returns current value of task which may be intermediate or final
//
// 3. If no args changed and existing task was done, flags task as COMPLETE
//
//============================================================================
void TaskList::UpdateTask(bool (* fn_ptr)(task *), cpp_xloper *InArray, short num_args, cpp_xloper &RetVal)
{
	if(!m_ThreadIsCreated || !m_ThreadIsRunning)
	{
		RetVal.SetToError(xlerrNA);
		return;
	}

	if(!fn_ptr || !InArray || num_args < 1)
	{
		RetVal.SetToError(xlerrValue);
		return;
	}

	xlName Caller;

	if(Caller.SetToCallersName()) // true if name exists
	{
// This cell has already a name assigned.  Need to check to see if
// this is an internal name associated with this function
		char *name = Caller.GetName();
		task *p = FindTask(name);
		free(name);

		if(p && fn_ptr == p->fn_ptr) // task exists already for this caller and function
		{
// Check if args have changed since task was scheduled
			if(p->args_changed(InArray, num_args))
			{
// Stop the task being or becoming current
				p->break_task = true;
				if(p == m_pCurrent) // m_pCurrent changes in background
					BreakCurrentTask(p); // waits for task to break
// Modify the task's arguments
				p->modify_args(InArray, num_args);
// Move the task to the end of the list.
				MoveTaskToTail(p);
// Enable the task to become current
				p->break_task = false;
			}

			if(p->status == TASK_UNCLAIMED)
				p->status = TASK_COMPLETE;

			RetVal = p->FnRetVal; // makes a deep copy
			return;
		}
//	fall through to MakeNewTask
	}

	if(!MakeNewTask(fn_ptr, InArray, num_args)) //, Caller))
		RetVal = false;
	else
		RetVal = true;
}
//============================================================================
bool TaskList::SetDoneTasks(void)
{
	char note_txt[50];
	double exec_secs;
	bool need_recalc = false;
	xlName R;
	task *p_temp;

	for(task *p = m_pHead; p;)
	{
		if(p->status != TASK_READY && p->status != TASK_COMPLETE)
		{
			p = p->next;
			continue;
		}

// Delete the task if its name no longer corresponds to a
// valid reference
		if(!R.Set(p->caller_name))
		{
			p_temp = p->next;
			DeleteTask(p); // frees all task resources
			p = p_temp;
			continue;
		}

// READY and COMPLETE tasks are flagged as unclaimed.
// This ensures that all orphaned tasks are deleted if, after
// recalculation, they have not been reset to TASK_COMPLETE
// through the action of a worksheet cell recalculation
		p->status = TASK_UNCLAIMED;

		exec_secs = (p->end_clock - p->start_clock) / (double)CLOCKS_PER_SEC;
		sprintf(note_txt, "Time: %.1lfs", exec_secs);
		R.SetNote(note_txt); // Put some information into the caller's note
		need_recalc = true;
		p = p->next;
	}
	return need_recalc;
}
//============================================================================
//============================================================================
//============================================================================
//
//	THESE FUNCTIONS ARE ONLY CALLED BY THE BACKGROUND THREAD
//
//============================================================================
//============================================================================
//============================================================================
// This function is not part of the class, but is passed as an argument
// to CreateThread() along with a pointer to 'this' so that it can call
// back into the class' TaskThreadMain. 
//============================================================================
DWORD __stdcall background_thread_main(void *vp)
{
	return ((TaskList *)vp)->TaskThreadMain();
}
//============================================================================
void TaskList::GetNextTask(void) // used by background to set m_pCurrent
{
	if(!m_pHead) // list is empty
	{
		m_pCurrent = NULL;
		return;
	}

	EnterCriticalSection(&m_ListControl);

	if(!m_pCurrent)
		m_pCurrent = m_pHead;

	do
	{
		if(m_pCurrent->status == TASK_PENDING && !m_pCurrent->break_task)
			break;

		m_pCurrent = m_pCurrent->next;
	}
	while(m_pCurrent); // look till end of list

	if(m_pCurrent)
		m_pCurrent->status = TASK_CURRENT;

	LeaveCriticalSection(&m_ListControl);
}
//============================================================================
DWORD TaskList::TaskThreadMain(void)
{
	for(;!m_ThreadExitFlagSet;)
	{
		if(!m_ThreadIsRunning)
		{
// Thread has been put into inactive state
			Sleep(THREAD_INACTIVE_SLEEP_MS);
			continue;
		}

		if(m_SuspendAllFlagSet)
		{
			m_ThreadIsRunning = false;
			m_pCurrent = NULL;
			continue;
		}

// Find next task to be executed. Sets m_pCurrent to
// point to the next task, or to NULL if no more to do.
		GetNextTask();

		if(m_pCurrent)
		{
// Execute the current task and time it. Status == TASK_CURRENT
			m_pCurrent->start_clock = clock();
			if(m_pCurrent->fn_ptr(m_pCurrent))
			{
// Task completed sucessfully and result is ready to be read out
				m_pCurrent->status = TASK_READY;
			}
			else
			{
// Task was broken or failed so need to re-queue it
				m_pCurrent->status = TASK_PENDING;
			}
			m_pCurrent->end_clock = clock();
		}
		else // nothing to do, so have a little rest
			Sleep(m_ThreadSleepMs);
	}
	return !(STILL_ACTIVE);
}
//============================================================================
//============================================================================
//============================================================================
//============================================================================
//
//  Example interface functions
//
//	These functions and commands work with the TaskList to perform
//	long tasks in background
//
//============================================================================
//============================================================================
//============================================================================
// LongTaskExampleMain() exectues the task and does the work.
// It is only ever called from the background thread.  It is
// required to check the break_task flag regularly to see if the
// foreground thread needs execution to stop.  It is not required
// that the task populates the return value, fn_ret_val, as it does
// in this case.  It could just wait till the final result is known.
//============================================================================
bool LongTaskExampleMain(task *pTask)
{
	long limit;

	if((limit = (long)(double)pTask->ArgArray[0]) < 1)
		return false;

	pTask->FnRetVal = 0.0;

	for(long i = 1; i <= limit; i++)
	{
		if(i % 1000)
		{
			if(pTask->break_task)
				return false;

			Sleep(0);
		}
		pTask->FnRetVal = (double)i;
	}
	return true;
}
//============================================================================
#define SECS_PER_DAY		(60.0 * 60.0 * 24.0)
//============================================================================
// This command function is called by Excel directly or indirectly
// in the foreground DLL thread.
//============================================================================
CmdRegData RC_LongTaskPollingCmd( // Instance name is not important
	"LongTaskPollingCmd",	// Name in code
	"LongTaskPollingCmd");	// Name that Excel uses

int __stdcall LongTaskPollingCmd(void)
{
#pragma EXPORT

	if(ExampleTaskList.m_BreakPollingCmdFlag)
		return 0; // return without rescheduling next call

// Run through the list of tasks setting TASK_READY tasks to
// TASK_UNCLAIMED. Tasks still unclaimed after recalculation are
// assumed to be orphaned and deleted by DeleteUnclaimedTasks().
	bool need_recalc = ExampleTaskList.SetDoneTasks();

	cpp_xloper Op; // Used to access Excel via the C API

//	if(need_recalc) // Commented out in this example
	{
// Cause Excel to recalculate.  This forces all volatile fns to be
// re-evaluated, including the long task functions, which will then
// return the most up-to-date values.  This also causes status of
// tasks to be changed to TASK_COMPLETE from TASK_UNCLAIMED.
		Op.Excel(xlcCalculateNow);

// Run through the list of tasks again to clean up unclaimed tasks
		ExampleTaskList.DeleteUnclaimedTasks();
	}

// Reschedule the command to repeat in m_PollingCmdFreqSecs seconds.
	Op.Excel(xlfNow); // Now as a serial time
	cpp_xloper ExecTime((double)Op +
			ExampleTaskList.GetPollingSecs() / SECS_PER_DAY);

// Use command name as given to Excel in xlfRegister 4th arg
	cpp_xloper CmdName("LongTaskPollingCmd"); // as registered with Excel

	if(Op.Excel(xlcOnTime, 2, &ExecTime, &CmdName) != xlretSuccess)
	{
		Op = "Can't reschedule long task polling cmd";
		Op.Alert(3);
	}
	return 1;
}
//============================================================================
// This command function can be called by Excel directly or indirectly
// in the foreground (main) Excel/DLL thread.
//============================================================================
CmdRegData RC_LongTaskToggleCmd( // Instance name is not important
	"LongTaskToggleCmd",	// Name in code
	"LongTaskToggleCmd");	// Name that Excel uses

int __stdcall LongTaskToggleCmd(void)
{
#pragma EXPORT

	if(!ExampleTaskList.IsThreadCreated())
		return 0;

	if(ExampleTaskList.IsThreadRunning())
		ExampleTaskList.SuspendTaskExecution();
	else
		ExampleTaskList.ResumeTaskExecution();

	return 1;
}
//============================================================================

// Long Task Thread Configuration Dialog

#define LTCD_NUM_ROWS	18

cpp_xloper LongTaskConfigDlg[LTCD_NUM_ROWS * CAPI_DLG_COLUMNS] =
{
	"", "", "", 490, 160, "Long Task Background Thread Configuration", "", // 
	1, 50, 120, 90, "", "OK", "", // Default OK button
	2, 150, 120, 90, "", "Cancel", "", // Cancel button
	3, 250, 120, "", "", "&Suspend", "", // OK button
	3, 350, 120, "", "", "&Kill thread", "", // OK button
	5, 20, 13, "", "", "", "", // Static text
	5, 20, 33, "", "", "Thread sleep time (ms)", "", // Static text
	5, 20, 53, "", "", "Command repeat time (s)", "", // Static text
	5, 20, 73, "", "", "Thread is initialised", "", // Static text
	5, 20, 93, "", "", "Thread is running", "", // Static text
	5, 300, 73, "", "", "Pending tasks:", "", // Static text
	5, 300, 93, "", "", "Done tasks:", "", // Static text
	5, 400, 73, "", "", "", "", // Static text
	5, 400, 93, "", "", "", "", // Static text
	7, 200, 30, 75, "", "", "", // Integer edit box
	7, 200, 50, 75, "", "", "", // Integer edit box
	213, 200, 73, "", "", "", "", // Check box
	213, 200, 93, "", "", "", "", // Check box
};

// Rows corresponding to the elements in the dialog box
enum
{
	LTCD_TOP,
	LTCD_OK,
	LTCD_CANCEL,
	LTCD_SUSP_REST,
	LTCD_KILL,
	LTCD_THREAD_ID_TXT,
	LTCD_THREAD_SLEEP_MS_TXT,
	LTCD_CMD_RPT_SECS_TXT,
	LTCD_IS_CREATED_TXT,
	LTCD_IS_RUNNING_TXT,
	LTCD_PENDING_TASKS_TXT,
	LTCD_DONE_TASKS_TXT,
	LTCD_PENDING_TASKS,
	LTCD_DONE_TASKS,
	LTCD_THREAD_SLEEP_MS,
	LTCD_CMD_RPT_SECS,
	LTCD_IS_CREATED,
	LTCD_IS_RUNNING,
};

//============================================================================
// This command function is called by Excel directly or indirectly
// in the foreground (main) Excel/DLL thread.
//============================================================================
CmdRegData RC_LongTaskConfigCmd( // Instance name is not important
	"LongTaskConfigCmd",	// Name in code
	"LongTaskConfigCmd");	// Name that Excel uses

int __stdcall LongTaskConfigCmd(void)
{
#pragma EXPORT

	if(!ExampleTaskList.IsThreadCreated())
		return 0;
	
// Create an array of xlopers (within a cpp_xloper) from the array of cpp_xlopers
	cpp_xloper DialogDef((RW)LTCD_NUM_ROWS, (COL)CAPI_DLG_COLUMNS, LongTaskConfigDlg);

// Set up the thread ID text field
	char id_text[30];
	sprintf(id_text, "Thread ID: 0x%08x", ExampleTaskList.GetThreadId());
	DialogDef.SetArrayElt((RW)LTCD_THREAD_ID_TXT, (COL)CAPI_DLG_TEXT, id_text);

	int item_num, temp;
	bool keep_looping = true;
	cpp_xloper Op, True(true), False(false);
	long pending, done;

	do
	{
// Set up the thread initialisation status field
		DialogDef.SetArrayElt((RW)LTCD_IS_CREATED, (COL)CAPI_DLG_VALUE, ExampleTaskList.IsThreadCreated());

// Set up the thread sleep ms and polling command repeat secs value fields
		DialogDef.SetArrayElt((RW)LTCD_THREAD_SLEEP_MS, (COL)CAPI_DLG_VALUE, (int)ExampleTaskList.GetThreadSleepMs());
		DialogDef.SetArrayElt((RW)LTCD_CMD_RPT_SECS, (COL)CAPI_DLG_VALUE, (int)ExampleTaskList.GetPollingSecs());

// Set the text on the suspend/restore button and the value of the checkbox
		DialogDef.SetArrayElt((RW)LTCD_SUSP_REST, (COL)CAPI_DLG_TEXT, ExampleTaskList.IsThreadRunning() ? "&Suspend" : "&Activate");
		DialogDef.SetArrayElt((RW)LTCD_IS_RUNNING, (COL)CAPI_DLG_VALUE, ExampleTaskList.IsThreadRunning());

// Set the text on the suspend/restore button and the value of the checkbox
		ExampleTaskList.GetTaskStats(pending, done);
		DialogDef.SetArrayElt((RW)LTCD_PENDING_TASKS, (COL)CAPI_DLG_TEXT, (int)pending);
		DialogDef.SetArrayElt((RW)LTCD_DONE_TASKS, (COL)CAPI_DLG_TEXT, (int)done);

// Restrict user input to the dialog and then display it
		Op.Excel(xlcDisableInput, 1, &True);

		if(Op.Excel(xlfDialogBox, 1, &DialogDef) != xlretSuccess || !ExampleTaskList.IsThreadCreated())
			break;

// Process the input from the dialog by reading the 7th column of the returned array.
// 7th column (CAPI_DLG_VALUE) of top row contains offset of selected button as a double.

		if(!Op.GetArrayElt((DWORD)CAPI_DLG_VALUE, item_num))
			break;

		switch(item_num)
		{
		case LTCD_SUSP_REST:
		// Toggle the thread state (and the text on the button) then fall
		// trough and check the values of the input fields
			if(ExampleTaskList.IsThreadRunning())
				ExampleTaskList.SuspendTaskExecution();
			else
				ExampleTaskList.ResumeTaskExecution();

		// Process the LTCD_THREAD_SLEEP_MS and LTCD_CMD_RPT_SECS fields
		case LTCD_OK:
			Op.GetArrayElt((RW)LTCD_THREAD_SLEEP_MS, (COL)CAPI_DLG_VALUE, temp);

			if(temp != ExampleTaskList.GetThreadSleepMs())
				ExampleTaskList.SetThreadSleepMs(temp);

			Op.GetArrayElt((RW)LTCD_CMD_RPT_SECS, (COL)CAPI_DLG_VALUE, temp);

			if(temp != ExampleTaskList.GetPollingSecs())
				ExampleTaskList.SetPollingSecs(temp);

			keep_looping = false;
			break;

		case LTCD_KILL:
		// Kill the thread and disable everything except the
		// OK button.  Redisplay the dialog one last time to
		// show the user that this has all been done.
			ExampleTaskList.DeleteTaskThread();

			for(WORD i = LTCD_CANCEL; i <= LTCD_CMD_RPT_SECS; i++)
			{
				DialogDef.GetArrayElt(i, (COL)CAPI_DLG_ITEM, item_num);
				item_num = 200 + (item_num % 100);
				DialogDef.SetArrayElt(i, (COL)CAPI_DLG_ITEM, item_num);
			}
			break;
		}
	}
	while(keep_looping);

// Re-enable user input to Excel
	Op.Excel(xlcDisableInput, 1, &False);
	return 1;
}
