//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================

//============================================================================
//============================================================================
// This source file contains the functions that Excel's Add-in Manager looks
// for as part of the XLL interface to the DLL's functionality.  It contains
// the code that registers the DLL's functions and commands and cleans-up when
// the XLL is being closed.  It contains the functions called by Excel to free
// DLL-allocated memory for things that were returned to it by the DLL during
// worksheet function calls.
//============================================================================
//============================================================================
#include <windows.h>
#include <stdio.h>

#include "XllAddIn.h"

HANDLE g_hInst = NULL;

//========================================================================
//========================================================================
#if STATIC_LINK_XLCALL32 == 0
//========================================================================
//========================================================================

// Declare function pointers that will be assigned at run-time using GetProcAddress()
pfnEXCEL4 Excel4 = NULL;
pfnEXCEL4v Excel4v = NULL;
pfnXLCALLVER XLCallVer = NULL;

#endif

// XL_AUTO_FREE_XLOPER determines whether or not xlAutoFree and
// xlAutoFree12 also free the passed-in xloper pointer as well as
// any memory pointed to within it.  Set this to non-zero if you return
// xlopers that are dynamically allocated using malloc or calloc.


#define XL_AUTO_FREE_XLOPER		0

#include "Background.h"

void customise_Excel_UI(bool set_up);

// Internal name list management function: cleans the list
int __stdcall delete_all_xll_names(void);

bool xll_initialised = false;

//========================================================================
#define CLASS_NAME_BUFFER_SIZE    50
#define WINDOW_TEXT_BUFFER_SIZE   50

// Data structure used as input to xldlg_enum_proc(), which is called by
// called_from_paste_fn_dlg(), called_from_replace_dlg(), and
// called_from_Excel_dlg().  These functions tell the caller whether
// the current worksheet function was called from one or either of
// these dialogs 

typedef struct
{
	bool is_dlg;
	short low_hwnd;
	char *window_title_text; // set to NULL if don't care
}
	xldlg_enum_struct;


//------------------------------------------------------------------------
// The callback function called by Windows for every top-level window
BOOL __stdcall xldlg_enum_proc(HWND hwnd, xldlg_enum_struct *p_enum)
{
#pragma EXPORT

// Check if the parent window is Excel
	if(LOWORD((DWORD)GetParent(hwnd)) != p_enum->low_hwnd)
		return TRUE; // keep iterating

	char class_name[CLASS_NAME_BUFFER_SIZE + 1];
//	Ensure that class_name is always null terminated
	class_name[CLASS_NAME_BUFFER_SIZE] = 0;

	GetClassName(hwnd, class_name, CLASS_NAME_BUFFER_SIZE);

//	Do a case-insensitve comparison for the Excel dialog window
//	class name with the Excel version number truncated
	if(_strnicmp(class_name, "bosa_sdm_xl", 11) == 0)
	{
// Check if a searching for a specific title string
		if(p_enum->window_title_text) 
		{
// Get the window's title and see if it matches the given text
			char buffer[WINDOW_TEXT_BUFFER_SIZE + 1];
			buffer[WINDOW_TEXT_BUFFER_SIZE] = 0;

// If title was "" and if we're looking for specific text
			if(GetWindowText(hwnd, buffer, WINDOW_TEXT_BUFFER_SIZE) == 0
			&& p_enum->window_title_text[0] != 0)
					return TRUE; // No match, so keep iterating

			if(buffer[0] != 0 && p_enum->window_title_text[0] != 0
			&& _stricmp(buffer, p_enum->window_title_text) != 0)
				return TRUE; // Keep iterating
		}
		p_enum->is_dlg = true;
		return FALSE;  // Tells Windows to stop iterating
	}
	return TRUE; // Tells Windows to continue iterating
}
//------------------------------------------------------------------------
bool called_from_paste_fn_dlg(void)
{
	xloper hwnd = {0.0, xltypeNil}; // super-safe

// Calls Excel4, so only returns the low part of Excel's
// main window handle.  This is ok for the search however.
	if(Excel4(xlGetHwnd, &hwnd, 0))
		return false; // Couldn't get it, so assume not

// Search for bosa_sdm_xl* dialog with no title string
	xldlg_enum_struct es = {FALSE, hwnd.val.w, ""};
	EnumWindows((WNDENUMPROC)xldlg_enum_proc, (LPARAM)&es);
	return es.is_dlg;
}
//------------------------------------------------------------------------
bool called_from_replace_dlg(void)
{
	xloper hwnd = {0.0, xltypeNil}; // super-safe

// Calls Excel4, so only returns the low part of Excel's
// main window handle.  This is ok for the search however.
	if(Excel4(xlGetHwnd, &hwnd, 0))
		return false; // Couldn't get it, so assume not

// Search for bosa_sdm_xl* dialog containing "Replace"
	xldlg_enum_struct es = {FALSE, hwnd.val.w, "Replace"};
	EnumWindows((WNDENUMPROC)xldlg_enum_proc, (LPARAM)&es);
	return es.is_dlg;
}
//------------------------------------------------------------------------
bool called_from_Excel_dlg(void)
{
	xloper hwnd = {0.0, xltypeNil}; // super-safe

// Calls Excel4, so only returns the low part of Excel's
// main window handle.  This is ok for the search however.
	if(Excel4(xlGetHwnd, &hwnd, 0))
		return false; // Couldn't get it, so assume not

// Search for bosa_sdm_xl* dialog (ignore title)
	xldlg_enum_struct es = {FALSE, hwnd.val.w, NULL};
	EnumWindows((WNDENUMPROC)xldlg_enum_proc, (LPARAM)&es);
	return es.is_dlg;
}
//========================================================================
// Data structure used as input to get_hwnd_enum_proc(), which is called by
// get_xl_main_handle(). This function retrieves the main Excel window handle

typedef struct
{
	short xl_low_handle;
	HWND full_handle;
}
	get_hwnd_enum_struct;

//------------------------------------------------------------------------
// The callback function called by Windows for every top-level window
BOOL __stdcall get_hwnd_enum_proc(HWND hwnd, get_hwnd_enum_struct *p_enum)
{
#pragma EXPORT

// Check if the low word of the handle matches Excel's
	if(LOWORD((DWORD)hwnd) != p_enum->xl_low_handle)
		return TRUE; // keep iterating

	char class_name[CLASS_NAME_BUFFER_SIZE + 1];
//	Ensure that class_name is always null terminated
	class_name[CLASS_NAME_BUFFER_SIZE] = 0;

	GetClassName(hwnd, class_name, CLASS_NAME_BUFFER_SIZE);

// Do a case-insensitive comparison for Excel's main window class name
	if(_stricmp(class_name, "xlmain") == 0)
	{
		p_enum->full_handle = hwnd;
		return FALSE;  // Tells Windows to stop iterating
	}
	return TRUE; // Tells Windows to continue iterating
}
//------------------------------------------------------------------------
HWND get_xl_main_handle(void)
{
	if(gExcelVersion12plus) // xlGetHwnd returns full handle
	{
		xloper12 main_xl_handle;
		if(Excel12(xlGetHwnd, &main_xl_handle, 0) != xlretSuccess)
			return 0;
		return (HWND)main_xl_handle.val.w;
	}
	else // xlGetHwnd returns low handle only
	{
		xloper main_xl_handle;
		if(Excel4(xlGetHwnd, &main_xl_handle, 0) != xlretSuccess)
			return 0;
		get_hwnd_enum_struct eproc_param = {main_xl_handle.val.w, 0};
		EnumWindows((WNDENUMPROC)get_hwnd_enum_proc, (LPARAM)&eproc_param);
		return eproc_param.full_handle;
	}
}
//============================================================================
// Can only call xlfGetWorkspace from a macro-sheet function or command, so if
// that fails, then is a worksheet function
bool caller_is_ws_function(void)
{
// Retrieve the R1C1 state. Fails if called from WS function
// Don't need the return value so pass NULL.
	if(gExcelVersion12plus)
	{
		xloper12 four;
		four.val.w = 4; // Ask if Excel is in R1C1 mode
		four.xltype = xltypeInt;
		return Excel12(xlfGetWorkspace, 0, 1, &four) != xlretSuccess;
	}
	else
	{
		xloper four;
		four.val.w = 4; // Ask if Excel is in R1C1 mode
		four.xltype = xltypeInt;
		return Excel4(xlfGetWorkspace, 0, 1, &four) != xlretSuccess;
	}
}

//============================================================================
int gExcelVersion = 0; // version not known
bool gExcelVersion12plus = false; // assumes xloper12s and the Excel12 interface not available
bool gExcelVersion11minus = true;
HMODULE hXLCall32dll = 0;


//============================================================================
#if STATIC_LINK_XLCALL32 == 0
bool link_Excel_API(void)
{
// First, check if the C API interface functions are defined.  If project
// was linked with an import library they should be, but if linking with
// xlcall32.dll at run-time, need to get the proc addresses for Excel4,
// Excel4v and XLCallVer.
	static bool already_failed = false;

	if(already_failed)
		return false;

	if(Excel4 == NULL)
	{
// Load the DLL and get the procedure addresses for Excel4 and Excel4v
// Module's handle is stored so can free the library in xlAutoClose
		hXLCall32dll = LoadLibrary("xlcall32.dll");
		if(!hXLCall32dll)
		{
			MessageBox(NULL, "Could not load xlcall32.dll",
				"Linking Excel API", MB_OK | MB_SETFOREGROUND);
			already_failed = true;
			return false;
		}
		Excel4 = (pfnEXCEL4)GetProcAddress(hXLCall32dll, "Excel4");
		Excel4v = (pfnEXCEL4v)GetProcAddress(hXLCall32dll, "Excel4v");
		XLCallVer = (pfnXLCALLVER)GetProcAddress(hXLCall32dll, "XLCallVer");

		if(!Excel4 || !Excel4v || !XLCallVer)
		{
			MessageBox(NULL, "Could not get addresses for Excel4, Excel4v and XLCallVer",
				"Linking Excel API", MB_OK | MB_SETFOREGROUND);
			already_failed = true;
			return false;
		}
	}
	return true;
}
//============================================================================
void unlink_Excel_API(void)
{
	if(hXLCall32dll)
	{
		FreeLibrary(hXLCall32dll);
		hXLCall32dll = 0;
		Excel4 = NULL;
		Excel4v = NULL;
		XLCallVer = NULL;
	}
}
#endif // STATIC_LINK_XLCALL32 == 0
//========================================================================
void set_global_Excel_version(void)
{
// Use xlopers, as we don't yet know if Excel 2007 and xloper12s are
// available.
	xloper version, arg;
	arg.val.w = 2;
	arg.xltype = xltypeInt;

	if(Excel4(xlfGetWorkspace, &version, 1, &arg) == xlretSuccess
	&& version.xltype == xltypeStr)
	{
		arg.val.w = xltypeInt;
// Convert version from string to integer (re-use arg for the return value)
		if(Excel4(xlCoerce, &arg, 2, &version, &arg) == xlretSuccess)
			gExcelVersion = arg.val.w;

		Excel4(xlFree, 0, 1, &version); // Free the Excel-allocated string
	}
	gExcelVersion12plus = (gExcelVersion >= 12);
	gExcelVersion11minus = !gExcelVersion12plus;
}
//========================================================================
int __stdcall xlAutoOpen(void)
{
#pragma EXPORT

	if(xll_initialised)
		return 1;

// Link to the C API.
#if STATIC_LINK_XLCALL32 == 0
	if(!link_Excel_API())
		return 1;
#endif

// Set the globally-accesible Excel version.
	set_global_Excel_version();

// Iterate through the instances of FnRegData (suports dual-verion interface) and CmdRegData
	FnRegData::RegisterAll();
	CmdRegData::RegisterAll();

// Setup any custom menus, event traps, etc.
	customise_Excel_UI(true);

// Initialise the background thread in the TaskList class
	ExampleTaskList.CreateTaskThread();

	xll_initialised = true;
	return 1;
}
//========================================================================
int __stdcall xlAutoClose(void)
{
#pragma EXPORT
	if(!xll_initialised)
		return 1;

// Explicitly clear the FnRegData and CmdRedData instance pointer tables
	FnRegData::ClearData();
	CmdRegData::ClearData();

// Remove the custom menus, event traps, etc.
	customise_Excel_UI(false);

// Shut down the background thread in the TaskList class
	ExampleTaskList.DeleteTaskThread();

// Delete all DLL-internal names including those created
// via instances of the xlName class
	delete_all_xll_names();

// Unlink the C API and reset the C API function pointers to NULL
#if STATIC_LINK_XLCALL32 == 0
	unlink_Excel_API();
#endif

	xll_initialised = false;
	return 1;
}
//========================================================================
xloper * __stdcall xlAutoRegister(xloper *p_name)
{
#pragma EXPORT
	return NULL; // Safe default action
}
//========================================================================
xloper12 * __stdcall xlAutoRegister12(xloper12 *p_name)
{
#pragma EXPORT
	return NULL; // Safe default action
}
//========================================================================
xloper * __stdcall xlAddInManagerInfo(xloper *p_arg)
{
#pragma EXPORT

	if(!xll_initialised)
		xlAutoOpen();

	if(!xll_initialised)
		return NULL;

	cpp_xloper Arg(p_arg);

	if(Arg != 1)
		return p_xlErrValue;

	cpp_xloper RetVal(AddinName);
	return RetVal.ExtractXloper();
}

xloper12 * __stdcall xlAddInManagerInfo12(xloper12 *p_arg)
{
#pragma EXPORT

	if(!xll_initialised)
		xlAutoOpen();

	if(!xll_initialised)
		return NULL;

	cpp_xloper Arg(p_arg);

	if(Arg != 1)
		return p_xl12ErrValue;

	cpp_xloper RetVal(AddinName);
	return RetVal.ExtractXloper12();
}
//========================================================================
void __stdcall xlAutoFree(xloper *p_oper)
{
#pragma EXPORT

	if(p_oper->xltype & xltypeMulti)
	{
// First check if string elements need to be freed then free the array
		int size = p_oper->val.array.rows * p_oper->val.array.columns;
		xloper *p = p_oper->val.array.lparray;

		for(; size-- > 0; p++) // check elements for strings
			if((p->xltype & ~(xlbitDLLFree | xlbitXLFree)) == xltypeStr)
			{
				if(p->xltype & xlbitDLLFree)
					free(p->val.str);
				else if(p->xltype & xlbitXLFree)
					Excel4(xlFree, 0, 1, p);
			}
		free(p_oper->val.array.lparray);
	}
	else if(p_oper->xltype & xltypeStr)
	{
		free(p_oper->val.str);
	}
	else if(p_oper->xltype & xltypeRef)
	{
		free(p_oper->val.mref.lpmref);
	}
#if XL_AUTO_FREE_XLOPER
	free(p_oper);
#endif
}
//========================================================================
void __stdcall xlAutoFree12(xloper12 *p_oper)
{
#pragma EXPORT

	if(p_oper->xltype & xltypeMulti)
	{
// First check if string elements need to be freed then free the array
		int size = p_oper->val.array.rows * p_oper->val.array.columns;
		xloper12 *p = p_oper->val.array.lparray;

		for(; size-- > 0; p++) // check elements for strings
			if((p->xltype & ~(xlbitDLLFree | xlbitXLFree)) == xltypeStr)
			{
				if(p->xltype & xlbitDLLFree)
					free(p->val.str);
				else if(p->xltype & xlbitXLFree)
					Excel12(xlFree, 0, 1, p);
			}
		free(p_oper->val.array.lparray);
	}
	else if(p_oper->xltype & xltypeStr)
	{
		free(p_oper->val.str);
	}
	else if(p_oper->xltype & xltypeRef)
	{
		free(p_oper->val.mref.lpmref);
	}
#if XL_AUTO_FREE_XLOPER
	free(p_oper);
#endif
}
//========================================================================
int __stdcall xlAutoAdd(void)
{
#pragma EXPORT

	if(!xll_initialised)
		xlAutoOpen();

	if(!xll_initialised)
		return 1;

	cpp_xloper xStr(AddinVersionStr);
	xStr.Alert();
	return 1;
}
//========================================================================
int __stdcall xlAutoRemove(void)
{
#pragma EXPORT

	if(!xll_initialised)
		return 1;

	cpp_xloper xStr("Add-in has been removed");
	xStr.Alert();
	return 1;
}
//========================================================================
// This function gets called when the DLL is first loaded and when
// detached, or when threads are attached or detached.
// Neither the C API nor COM should be called from this function.

// This implementation just calls a function that sets up thread-local storage
BOOL TLS_Action(DWORD Reason);

__declspec(dllexport) BOOL __stdcall DllMain(HINSTANCE hDll, DWORD Reason, void *Reserved)
{
	g_hInst = hDll;
	return TLS_Action(Reason);
}
