//============================================================================
//============================================================================
//
// Financial Applications of Excel Add-in Development in C/C++ (2nd edition)
// 
// Author: Steve Dalton
// 
// Published by John Wiley & Sons Ltd, The Atrium, Southern Gate, Chichester,
// West Sussex, PO19 8SQ, UK.
// 
// All intellectual property rights and copyright in the code listed in this
// module are reserved.  You may reproduce and use this code only as permitted
// in the Rules of Use that you agreed to on installation.  To use the material
// in this source module, or material derived from it, in any other way you
// must first obtain written permission.  Requests for permission can be sent
// by email to permissions@eigensys.com.
// 
// No warranty, explicit or implied, is made by either the author or publisher
// as to the quality, fitness for a particular purpose, accuracy or
// appropriateness of material in this module.  The code and methods contained
// are intended soley for example and clarification.  You should not rely on
// any part of this code without having completely satisfied yourself that it
// is correct and appropriate for your needs.
//
//============================================================================
//============================================================================

#include <windows.h>
#include <stdio.h>

#include "XllAddIn.h"

size_t count_char(char *text, char ch);
size_t count_chars(char *text, char *chars);
char *strspnp(char *text, char *search_text);

//========================================================================
void display_register_error(char *fn_name, int XL4_err_num, int err_num)
{
	char temp_buffer[256];
	_snprintf(temp_buffer, 256, "Could not register function %s (XL4:%d,Err:%d)", fn_name, XL4_err_num, err_num);
	cpp_xloper xStr(temp_buffer);
	xStr.Alert(3); // Error alert dialog type
}
//========================================================================
// Do some (limited) integrity-checking of the fn registration arguments
//========================================================================
#define V12_NEW_CHARS	"%QU$" // New characters introduced in Excel 12

bool check_register_args(char *fn_name, char *arg_types, char *arg_labels)
{
// Check for forbidden characters in the type string and count the args.
// %,Q,U,$ introduced in Excel 12 and CANNOT be passed in earlier versions.

	char *ok_1st_chars, *ok_other_chars, *pch = NULL;

	if(gExcelVersion12plus)
	{
		ok_1st_chars = "ABCDEFGHIJKLMNPQRU>123456789";
		ok_other_chars = "ABCDEFGHIJKLMNOPQRU!#$%";
	}
	else
	{
		ok_1st_chars = "ABCDEFGHIJKLMNPR>123456789";
		ok_other_chars = "ABCDEFGHIJKLMNOPR!#";
	}

// If 1st char is not one of the permitted characters, point pch to it
	if(!strchr(ok_1st_chars, arg_types[0]))
		pch = arg_types;
	else // check the others
		pch = strspnp(arg_types + 1, ok_other_chars);

	if(pch) // found something that shouldn't be there
	{
// Don't display error if simply Excel12 typed fn and we're running in Excel 11-
		if(gExcelVersion11minus && strchr(V12_NEW_CHARS, *pch) != NULL)
			return false;

		char problem_char[2] = {*pch, 0};
		cpp_xloper ErrMsg(fn_name);
		ErrMsg += ": Arg type contains invalid character: ";
		ErrMsg += problem_char;
		ErrMsg.Alert(3); // Error alert dialog type
		return false;
	}

// Count the number of arg type characters.  In Excel 12+, need to ignore
// %, i.e. K% is one argument, along with ignoring !, # and $
	size_t num_args = strlen(arg_types) - count_chars(arg_types, "!#$%") - 1;

// Number of args should be number of commas + 1
	if((num_args == 0 && *arg_labels != 0)
	|| (num_args != 0 && num_args != count_char(arg_labels, ',') + 1))
	{
		cpp_xloper ErrMsg("Register function args invalid: ");
		ErrMsg += fn_name;
		ErrMsg.Alert(3); // Error alert dialog type
		return false;
	}
	return true;
}
//========================================================================
//========================================================================
//========================================================================
//========================================================================
//========================================================================
long FnRegData::count = 0;
long FnRegData::array_size = 0;
FnRegData **FnRegData::InstanceArray = NULL;


FnRegData::FnRegData(char *name_in_code, char *types, char *name_in_code12,
	char *types12, char *ws_name, char *arg_names, char *arg_names12,
	char *fn_category, char *help_file, char *fn_description, ...)
{
	m_NameInCode = name_in_code;
	m_Types = types;
	m_NameInCode12 = name_in_code12;
	m_Types12 = types12;
	m_WsName = ws_name;
	m_ArgNames = arg_names;
	m_ArgNames12 = arg_names12;
	m_FnCategory = fn_category;
	m_HelpFile = help_file;
	m_FnDescription = fn_description;

	va_list arg_ptr;
	va_start(arg_ptr, fn_description); // Initialize

	int i;
	for(i = 0; i < MAX_XL12_UDF_ARGS - 11; i++)
		if((m_ArgHelp[i] = va_arg(arg_ptr, char *)) == NULL)
			break;

	va_end(arg_ptr); // Reset

	for(i++; i < MAX_XL12_UDF_ARGS - 11; i++)
		m_ArgHelp[i] = NULL;

	if(count == array_size) // then need to allocate more space
	{
		FnRegData **new_array = new FnRegData *[array_size + 20];
		if(array_size)
			memcpy(new_array, InstanceArray,
				array_size * sizeof(FnRegData *));
		array_size += 20;
		delete[] InstanceArray;
		InstanceArray = new_array;
	}
	InstanceArray[count++] = this;
}
//========================================================================
void FnRegData::RegisterAll(void)
{
	for(long i = 0; i < count; i++)
		InstanceArray[i]->RegisterThis();
}
//========================================================================
bool FnRegData::RegisterThis(void)
{
// These depend on the the version and what is provided
	char *types, *code_name, *arg_names;

	if(gExcelVersion12plus)
	{
		if(m_NameInCode12 && m_NameInCode12[0])
		{
			code_name = m_NameInCode12;
			types = m_Types12;
		}
		else
		{
			code_name = m_NameInCode;
			types = m_Types;
		}

		if(m_ArgNames12 && m_ArgNames12[0])
			arg_names = m_ArgNames12;
		else
			arg_names = m_ArgNames;
	}
	else
	{
		code_name = m_NameInCode;
		types = m_Types;
		arg_names = m_ArgNames;
	}

	if(!check_register_args(m_WsName, types, arg_names))
		return false;

	int arg_limit = (gExcelVersion12plus ? MAX_XL12_UDF_ARGS : MAX_XL11_UDF_ARGS);

// Array of pointers to xloper that will be passed to Excel4v()
	const cpp_xloper *ptr_array[MAX_XL12_UDF_ARGS];
	cpp_xloper FnArgs[MAX_XL12_UDF_ARGS];

// Get the full path and name of the DLL.
// Passed as the first argument to xlfRegister.
	if(FnArgs[0].Excel(xlGetName))
	{
		cpp_xloper ErrMsg = "Could not get XLL path,name";
		ErrMsg.Alert(3);// Error alert type
		return false;
	}

	FnArgs[1] = code_name;
	FnArgs[2] = types;
	FnArgs[3] = m_WsName;
	FnArgs[4] = arg_names;
	FnArgs[5] = 1;
	FnArgs[6] = m_FnCategory;
	FnArgs[7].SetType(xltypeMissing); // Short cut text character
	FnArgs[8] = m_HelpFile;
	FnArgs[9] = m_FnDescription;

	char *p_arg;
	int i;
	for(i = 10; i < arg_limit; i++)
	{
		p_arg = m_ArgHelp[i-10];
		if(!(p_arg && *p_arg))
			break; // that was the last of the arguments for this fn

// Set the corresponding xlfRegister argument
		FnArgs[i] = p_arg; // convert the string to a cpp_xloper
	}

// Set up the array of pointers
	int num_args;
	for(num_args = i; --i >= 0;)
		ptr_array[i] = FnArgs + i;

	cpp_xloper Op;
	if(Op.Excel(xlfRegister, num_args, ptr_array) != xlretSuccess)
	{
		Op = "Couldn't register ";
		Op += FnArgs[3];
		Op.Alert(3); // Error alert dialog type
		return false;
	}
	return true;
}
//========================================================================
//========================================================================
//========================================================================
//========================================================================
//========================================================================

long CmdRegData::count = 0;
long CmdRegData::array_size = 0;
CmdRegData **CmdRegData::InstanceArray = NULL;

CmdRegData::CmdRegData(char *name_in_code, char *macro_name)
{
	m_NameInCode = name_in_code;
	m_MacroName = macro_name;

	if(count == array_size) // then need to allocate more space
	{
		CmdRegData **new_array = new CmdRegData *[array_size + 20];
		if(array_size)
			memcpy(new_array, InstanceArray, array_size * sizeof(CmdRegData *));
		array_size += 20;
		delete[] InstanceArray;
		InstanceArray = new_array;
	}
	InstanceArray[count++] = this;
}
//========================================================================
void CmdRegData::RegisterAll(void)
{
	for(long i = 0; i < count; i++)
		InstanceArray[i]->RegisterThis();
}
//========================================================================
bool CmdRegData::RegisterThis(void)
{
	cpp_xloper RetVal, DllName;
//-------------------------------------------------------
// Get the full path and name of the DLL.
// Passed as the first argument to xlfRegister, so need
// to set first pointer in array to point to this.
//-------------------------------------------------------
	if(DllName.Excel(xlGetName) != xlretSuccess)
		return false;

//-------------------------------------------------------
// Set up the rest of the arguments.
//-------------------------------------------------------
	cpp_xloper CodeName(m_NameInCode);
	cpp_xloper ExcelName(m_MacroName);
	cpp_xloper RtnType("J");
	cpp_xloper MacroType(2); // Command
	cpp_xloper NilArgs; // defaults to xltypeNil
	int xl_ret_val = RetVal.Excel(xlfRegister, 6, &DllName,
		&CodeName, &RtnType, &ExcelName, &NilArgs, &MacroType);

	if(xl_ret_val != xlretSuccess)
		display_register_error(m_NameInCode, xl_ret_val, (int)RetVal);

	return true;
}
//========================================================================
//========================================================================
//========================================================================
//========================================================================
//========================================================================
